var __defProp = Object.defineProperty;
var __markAsModule = (target) => __defProp(target, "__esModule", {value: true});
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, {get: all[name], enumerable: true});
};

// program.imba
__markAsModule(exports);
__export(exports, {
  CompletionTypes: () => CompletionTypes,
  Group: () => Group,
  ImbaDocument: () => ImbaDocument,
  KeywordTypes: () => KeywordTypes,
  Keywords: () => Keywords,
  M: () => M,
  Monarch: () => Monarch,
  Node: () => Node,
  Position: () => Position,
  Range: () => Range,
  Scope: () => Scope,
  SemanticTokenModifiers: () => SemanticTokenModifiers,
  SemanticTokenTypes: () => SemanticTokenTypes,
  Sym: () => Sym,
  SymbolFlags: () => SymbolFlags,
  SymbolKind: () => SymbolKind,
  Token: () => Token,
  grammar: () => grammar,
  highlight: () => highlight,
  lexer: () => lexer
});

// src/program/structures.imba
var φ1 = Symbol.for("#init");
var φ2 = Symbol.for("#source");
var φ3 = Symbol.for("#lineText");
var φ4 = Symbol.for("#version");
var DOCMAP = new WeakMap();
var Position = class {
  [φ1]($$ = null) {
    this.line = $$ ? $$.line : void 0;
    this.character = $$ ? $$.character : void 0;
    this.offset = $$ ? $$.offset : void 0;
  }
  constructor(l, c, o, v = null) {
    this[φ1]();
    this.line = l;
    this.character = c;
    this.offset = o;
    this[φ4] = v;
  }
  toString() {
    return "" + this.line + ":" + this.character;
  }
  valueOf() {
    return this.offset;
  }
};
var Range = class {
  [φ1]($$ = null) {
    this.start = $$ ? $$.start : void 0;
    this.end = $$ ? $$.end : void 0;
  }
  constructor(start, end) {
    this[φ1]();
    this.start = start;
    this.end = end;
  }
  get offset() {
    return this.start.offset;
  }
  get length() {
    return this.end.offset - this.start.offset;
  }
  get ["0"]() {
    return this.start.offset;
  }
  get ["1"]() {
    return this.end.offset;
  }
  getText(str) {
    return str.slice(this.start, this.end);
  }
  equals(other) {
    return other.offset == this.offset && other.length == this.length;
  }
};

// src/program/utils.imba
function iter$__(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
function prevToken(start, pattern, max = 1e5) {
  let tok = start;
  while (tok && max > 0) {
    if (tok.match(pattern)) {
      return tok;
    }
    ;
    max--;
    tok = tok.prev;
  }
  ;
  return null;
}
function computeLineOffsets(text, isAtLineStart, textOffset) {
  if (textOffset === void 0) {
    textOffset = 0;
  }
  ;
  var result = isAtLineStart ? [textOffset] : [];
  var i = 0;
  while (i < text.length) {
    var ch = text.charCodeAt(i);
    if (ch === 13 || ch === 10) {
      if (ch === 13 && i + 1 < text.length && text.charCodeAt(i + 1) === 10) {
        i++;
      }
      ;
      result.push(textOffset + i + 1);
    }
    ;
    i++;
  }
  ;
  return result;
}
function getWellformedRange(range) {
  var start = range.start;
  var end = range.end;
  if (start.line > end.line || start.line === end.line && start.character > end.character) {
    return new Range(end, start);
  }
  ;
  return range instanceof Range ? range : new Range(start, end);
}
function editIsFull(e) {
  return e !== void 0 && e !== null && typeof e.text === "string" && e.range === void 0;
}
function fastExtractSymbols(text) {
  let lines = text.split(/\n/);
  let symbols = [];
  let scope = {indent: -1, children: []};
  let root = scope;
  let m;
  let t0 = Date.now();
  for (let i = 0, φ42 = iter$__(lines), φ52 = φ42.length; i < φ52; i++) {
    let line = φ42[i];
    if (line.match(/^\s*$/)) {
      continue;
    }
    ;
    let indent = line.match(/^\t*/)[0].length;
    while (scope.indent >= indent) {
      scope = scope.parent || root;
    }
    ;
    m = line.match(/^(\t*((?:export )?(?:static )?(?:extend )?)(class|tag|def|get|set|prop|attr) )(\@?[\w\-\$\:]+(?:\.[\w\-\$]+)?)/);
    if (m) {
      let kind = m[3];
      let name = m[4];
      let ns = scope.name ? scope.name + "." : "";
      let mods = m[2].trim().split(/\s+/);
      let md = "";
      let span = {
        start: {line: i, character: m[1].length},
        end: {line: i, character: m[0].length}
      };
      let symbol = {
        kind,
        ownName: name,
        name: ns + name,
        span,
        indent,
        modifiers: mods,
        children: [],
        parent: scope == root ? null : scope,
        type: kind,
        data: {},
        static: mods.indexOf("static") >= 0,
        extends: mods.indexOf("extend") >= 0
      };
      if (symbol.static) {
        symbol.containerName = "static";
      }
      ;
      symbol.containerName = m[2] + m[3];
      if (kind == "tag" && (m = line.match(/\<\s+([\w\-\$\:]+(?:\.[\w\-\$]+)?)/))) {
        symbol.superclass = m[1];
      }
      ;
      if (scope.type == "tag") {
        md = "```html\n<" + scope.name + " " + name + ">\n```\n";
        symbol.description = {kind: "markdown", value: md};
      }
      ;
      scope.children.push(symbol);
      scope = symbol;
      symbols.push(symbol);
    }
    ;
  }
  ;
  root.all = symbols;
  console.log("fast outline", text.length, Date.now() - t0);
  return root;
}

// src/program/grammar.imba
function iter$__2(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
var eolpop = [/^/, {token: "@rematch", next: "@pop"}];
var repop = {token: "@rematch", next: "@pop"};
var toodeep = {token: "white.indent", next: "@>illegal_indent"};
function regexify(array, pattern = "#") {
  if (typeof array == "string") {
    array = array.split(" ");
  }
  ;
  let items = array.slice().sort(function(_0, _1) {
    return _1.length - _0.length;
  });
  items = items.map(function(item) {
    let escaped = item.replace(/[.*+\-?^${}()|[\]\\]/g, "\\$&");
    return pattern.replace("#", escaped);
  });
  return new RegExp("(?:" + items.join("|") + ")");
}
function denter(indent, outdent, stay, o = {}) {
  var φ16;
  if (indent == null) {
    indent = toodeep;
  } else if (indent == 1) {
    indent = {next: "@>"};
  } else if (indent == 2) {
    indent = {next: "@>_indent&-_indent"};
  } else if (typeof indent == "string") {
    indent = {next: indent};
  }
  ;
  if (outdent == -1) {
    outdent = repop;
  }
  ;
  if (stay == -1) {
    stay = repop;
  } else if (stay == 0) {
    o.comment == null ? o.comment = true : o.comment;
    stay = {};
  }
  ;
  indent = Object.assign({token: "white.tabs"}, indent || {});
  stay = Object.assign({token: "white.tabs"}, stay || {});
  outdent = Object.assign({token: "@rematch", next: "@pop"}, outdent || {});
  let cases = {
    "$1==$S2	": indent,
    "$1==$S2": {
      cases: {"$1==$S6": stay, "@default": {token: "@rematch", switchTo: "@*$1"}}
    },
    "@default": outdent
  };
  φ16 = 0;
  for (let k of ["next", "switchTo"]) {
    let v = φ16++;
    if (indent[k] && indent[k].indexOf("*") == -1) {
      indent[k] += "*$1";
    }
    ;
  }
  ;
  let rule = [/^(\t*)(?=[^\t\n])/, {cases}];
  if (o.comment) {
    let clones = {};
    for (let φ23 = 0, φ32 = Object.keys(cases), φ42 = φ32.length, k, v; φ23 < φ42; φ23++) {
      k = φ32[φ23];
      v = cases[k];
      let clone = Object.assign({}, v);
      if (!clone.next && !clone.switchTo) {
        clone.next = "@>_comment";
      }
      ;
      clones[k] = clone;
    }
    ;
    return [[/^(\t*)(?=#\s|#$)/, {cases: clones}], rule];
  }
  ;
  return rule;
}
var states = {
  root: [
    [/^@comment/, "comment", "@>_comment"],
    [/^(\t+)(?=[^\t\n])/, {cases: {
      "$1==$S2	": {token: "white.indent", next: "@>_indent*$1"},
      "@default": "white.indent"
    }}],
    "block_"
  ],
  _comment: [
    [/^([\t\s\n]*)$/, "comment"],
    [/^(\t*)([\S\s]*)/, {cases: {
      "$1~$S2	*": {token: "comment"},
      "@default": {token: "@rematch", next: "@pop"}
    }}],
    [/[\S\s]+/, "comment"]
  ],
  illegal_indent: [
    denter()
  ],
  identifier_: [
    [/\$\w+\$/, "identifier.env"],
    [/\$\d+/, "identifier.special"],
    [/\#+@id/, "identifier.symbol"],
    [/\¶@id/, "ivar"],
    [/@id\!?/, {cases: {
      this: "this",
      self: "self",
      "@keywords": "keyword.$#",
      "$0~[A-Z].*": "identifier.uppercase.$F",
      "@default": "identifier.$F"
    }}]
  ],
  block_: [
    [/^(\t+)(?=[\r\n]|$)/, "white.tabs"],
    "class_",
    "tagclass_",
    "var_",
    "func_",
    "import_",
    "export_",
    "flow_",
    "for_",
    "try_",
    "catch_",
    "while_",
    "css_",
    "tag_",
    "do_",
    "block_comment_",
    "expr_",
    [/[ ]+/, "white"],
    "common_"
  ],
  indentable_: [
    denter("@>_paren_indent&-_indent", null, null),
    [/^(\t+)(?=[\r\n]|$)/, "white.tabs"],
    "whitespace"
  ],
  _indent: [
    denter(2, -1, 0),
    "block_"
  ],
  _paren_indent: [
    denter(2, -1, 0),
    "block_",
    [/\)|\}|\]/, "@rematch", "@pop"]
  ],
  block: [
    denter("@>", -1, 0),
    "block_"
  ],
  bool_: [
    [/(true|false|yes|no|undefined|null)(?![\:\-\w\.\_])/, "boolean"]
  ],
  op_: [
    [/\s+\:\s+/, "operator.ternary"],
    [/(@unspaced_ops)/, {cases: {
      "@spread": "operator.spread",
      "@access": "operator.access",
      "@default": "operator"
    }}],
    [/\/(?!\/)/, "operator.math"],
    [/\&(?=[,\)])/, "operator.special.blockparam"],
    [/(\s*)(@symbols)(\s*)/, {cases: {
      "$2@assignments": "operator.assign",
      "$2@math": "operator.math",
      "$2@operators": "operator",
      "$2@logic": "operator.logic",
      "$2@access": "operator.access",
      "@default": "delimiter"
    }}],
    [/\&\b/, "operator"]
  ],
  keyword_: [
    [/new@B/, "keyword.new"],
    [/isa@B/, "keyword.isa"],
    [/is@B/, "keyword.is"],
    [/(switch|when|throw|continue|break|then|await|typeof|by)@B/, "keyword.$1"],
    [/delete@B/, "keyword.delete"],
    [/and@B|or@B/, "operator.flow"]
  ],
  return_: [
    [/return@B/, "keyword.new"]
  ],
  primitive_: [
    "string_",
    "number_",
    "regexp_",
    "bool_"
  ],
  value_: [
    "primitive_",
    "keyword_",
    "implicit_call_",
    "parens_",
    "key_",
    "access_",
    "identifier_",
    "array_",
    "object_"
  ],
  expr_: [
    "comment_",
    "inline_var_",
    "return_",
    "value_",
    "tag_",
    "op_",
    "type_",
    "spread_"
  ],
  attr_expr_: [
    "primitive_",
    "parens_",
    "access_",
    "identifier_",
    "array_",
    "object_",
    "tag_",
    "op_"
  ],
  access_: [
    [/(\.\.?)(@propid\!?)/, {cases: {
      "$2~[A-Z].*": ["operator.access", "accessor.uppercase"],
      "$2~#.*": ["operator.access", "accessor.symbol"],
      "@default": ["operator.access", "accessor"]
    }}]
  ],
  call_: [
    [/\(/, "(", "@call_body"]
  ],
  key_: [
    [/(\#+@id)(\:\s*)/, ["key.symbol", "operator.assign.key-value"]],
    [/(@propid)(\:\s*)/, {cases: {
      "@default": ["key", "operator.assign.key-value"]
    }}]
  ],
  implicit_call_: [
    [/(\.\.?)(@propid)@implicitCall/, {cases: {
      "$2~[A-Z].*": ["operator.access", "accessor.uppercase", "@implicit_call_body"],
      "@default": ["operator.access", "accessor", "@implicit_call_body"]
    }}],
    [/(@propid)@implicitCall/, {cases: {
      "$2~[A-Z].*": ["identifier.uppercase", "@implicit_call_body"],
      "@default": ["identifier", "@implicit_call_body"]
    }}]
  ],
  implicit_call_body: [
    eolpop,
    [/\)|\}|\]|\>/, "@rematch", "@pop"],
    "arglist_",
    "whitespace"
  ],
  arglist_: [
    "do_",
    "expr_",
    [/\s*\,\s*/, "delimiter.comma"]
  ],
  params_: [
    [/\[/, "array.[", "@array_var_body=decl-param"],
    [/\{/, "object.{", "@object_body=decl-param"],
    [/(@variable)/, "identifier.decl-param"],
    "spread_",
    "type_",
    [/\s*\=\s*/, "operator", "@var_value="],
    [/\s*\,\s*/, "separator"]
  ],
  object_: [
    [/\{/, "object.{", "@object_body"]
  ],
  parens_: [
    [/\(/, "parens.(", "@parens_body"]
  ],
  parens_body: [
    [/\)/, ")", "@pop"],
    "indentable_",
    "arglist_"
  ],
  array_: [
    [/\[/, "array.[", "@array_body"]
  ],
  array_body: [
    [/\]@implicitCall/, {token: "]", switchTo: "@implicit_call_body="}],
    [/\]/, "]", "@pop"],
    "indentable_",
    "expr_",
    [",", "delimiter"]
  ],
  object_body: [
    [/\}/, "}", "@pop"],
    [/(@id)(\s*:\s*)/, ["key", "operator.assign.key-value", "@object_value"]],
    [/(@id)/, "identifier.$F"],
    [/\[/, "[", "@object_dynamic_key="],
    [/\s*=\s*/, "operator.assign", "@object_value="],
    [/:/, "operator.assign.key-value", "@object_value="],
    [/\,/, "delimiter.comma"],
    "indentable_",
    "expr_"
  ],
  object_value: [
    eolpop,
    [/,|\}|\]|\)/, "@rematch", "@pop"],
    "expr_"
  ],
  object_dynamic_key: [
    ["]", "]", "@pop"],
    "expr_"
  ],
  comment_: [
    [/#(\s.*)?(\n|$)/, "comment"]
  ],
  block_comment_: [
    [/###/, "comment.start", "@_block_comment"]
  ],
  _block_comment: [
    [/###/, "comment.end", "@pop"],
    [/[^#]+/, "comment"],
    [/#(?!##)/, "comment"]
  ],
  try_: [
    [/try@B/, "keyword.try", "@>_try&try"]
  ],
  catch_: [
    [/(catch\s+)(?=@id(\s|$))/, "keyword.catch", "@catch_start&catch"],
    [/catch@B/, "keyword.catch", "@catch_start&catch"]
  ],
  catch_start: [
    [/@id/, "identifier.decl-const", {switchTo: "@>_catch"}],
    [/.?/, "@rematch", {switchTo: "@>_catch"}]
  ],
  _catch: [
    denter("@>block", -1, 0),
    "block_"
  ],
  _try: [
    denter("@>block", -1, 0),
    "block_"
  ],
  do_: [
    [/do(?=\()/, "keyword.do", "@>do_start&do"],
    [/do@B/, "keyword.do", "@>_do&do"]
  ],
  do_start: [
    denter(null, -1, -1),
    [/\(/, "(", {switchTo: "@_do_params"}],
    [/./, "@rematch", {switchTo: "@_do"}]
  ],
  _do_params: [
    [/\)/, ")", {switchTo: "@_do"}],
    "params_"
  ],
  _do: [
    denter(2, -1, 0),
    [/(\}|\)|\])/, "@rematch", "@pop"],
    "block_"
  ],
  class_: [
    [/(extend)(?=\s+class )/, "keyword.$1"],
    [/(global)(?=\s+class )/, "keyword.$1"],
    [/(class)(\s)(@id)/, ["keyword.$1", "white.$1name", "entity.name.class.decl-const", "@class_start="]],
    [/(class)(?=\n)/, "keyword.$1", "@>_class&class="]
  ],
  class_start: [
    [/(\s+\<\s+)(@id)/, ["keyword.extends", "identifier.superclass"]],
    [/@comment/, "comment"],
    [/^/, "@rematch", {switchTo: "@>_class&class="}]
  ],
  tagclass_: [
    [/(extend)(?=\s+tag )/, "keyword.$1"],
    [/(global)(?=\s+tag )/, "keyword.$1"],
    [/(tag)(\s)(@constant)/, ["keyword.tag", "white.tagname", "entity.name.component.local", "@tagclass_start="]],
    [/(tag)(\s)(@id)/, ["keyword.tag", "white.tagname", "entity.name.component.global", "@tagclass_start="]]
  ],
  tagclass_start: [
    [/(\s+\<\s+)(@id)/, ["keyword.extends", "identifier.superclass"]],
    [/@comment/, "comment"],
    [/^/, "@rematch", {switchTo: "@>_tagclass&component="}]
  ],
  import_: [
    [/(import)(?=\s+['"])/, "keyword.import", "@>import_source"],
    [/(import)(\s+type)(?=\s[\w\$\@\{])/, ["keyword.import", "keyword.type", "@>import_body&-_imports=decl-import/part"]],
    [/(import)@B/, "keyword.import", "@>import_body&-_imports=decl-import/part"]
  ],
  import_body: [
    denter(null, -1, 0),
    [/(@esmIdentifier)( +from)/, ["identifier.$F.default", "keyword.from", {switchTo: "@import_source"}]],
    [/(\*)(\s+as\s+)(@esmIdentifier)(\s+from)/, ["keyword.star", "keyword.as", "identifier.$F.ns", "keyword.from", {switchTo: "@import_source"}]],
    [/(@esmIdentifier)(\s*,\s*)(\*)(\s+as\s+)(@esmIdentifier)(from)/, ["identifier.$F.default", "delimiter.comma", "keyword.star", "keyword.as", "identifier.$F.ns", "keyword.from", {switchTo: "@import_source"}]],
    [/\ *from/, "keyword.from", {switchTo: "@import_source"}],
    [/\{/, "specifiers.{", "@esm_specifiers/part"],
    [/(@esmIdentifier)/, "identifier.$F", {switchTo: "@/delim"}],
    [/\s*\,\s*/, "delimiter.comma", {switchTo: "@/part"}],
    "comma_",
    "common_"
  ],
  import_source: [
    denter(null, -1, 0),
    [/["']/, "path.open", "@_path=$#"],
    eolpop
  ],
  export_: [
    [/(export)( +)(default)@B/, ["keyword.export", "white", "keyword.default"]],
    [/(export)(?= +(let|const|var|class|tag)@B)/, "keyword.export"],
    [/(export)( +)(global)@B/, ["keyword.export", "white", "keyword.global"]],
    [/(export)(\s+\*\s+)(from)@B/, ["keyword.export", "operator.star", "keyword.from", "@>import_source"]],
    [/(export)@B/, "keyword.export", "@>export_body"]
  ],
  export_body: [
    denter(null, -1, 0),
    [/(\*)(\s+as\s+)(@esmIdentifier)/, ["keyword.star", "keyword.as", "identifier.const.export"]],
    [/(@esmIdentifier)(\s+as\s+)(default)/, ["alias", "keyword.as", "alias.default"]],
    [/(@esmIdentifier)(\s+as\s+)(@esmIdentifier)/, ["alias", "keyword.as", "identifier.const.export"]],
    [/from/, "keyword.from", {switchTo: "@import_source"}],
    [/\{/, "{", "@esm_specifiers=export/part"],
    [/(@esmIdentifier)/, "identifier.const.export"],
    [/\*/, "operator.star"],
    "comma_",
    "common_"
  ],
  esm_specifiers: [
    [/\}/, "}", "@pop"],
    [/(@esmIdentifier)(\s+as\s+)(@esmIdentifier)/, ["alias", "keyword.as", "identifier.const.$F", {switchTo: "@/delim"}]],
    [/@esmIdentifier/, {cases: {
      "$/==part": {token: "identifier.const.$S4", switchTo: "@/delim"},
      "@default": {token: "invalid"}
    }}],
    [/\s*\,\s*/, "delimiter.comma", {switchTo: "@/part"}],
    "whitespace"
  ],
  _path: [
    [/[^"'\`\{\\]+/, "path"],
    [/@escapes/, "path.escape"],
    [/\./, "path.escape.invalid"],
    [/\{/, "invalid"],
    [/["'`]/, {cases: {"$#==$F": {token: "path.close", next: "@pop"}, "@default": "path"}}]
  ],
  member_: [
    [/(constructor)@B/, "entity.name.constructor", "@>def_params&def/def"],
    [/(def|get|set)(\s)(@defid)/, ["keyword.$1", "white.entity", "entity.name.$1", "@>def_params&$1/$1"]],
    [/(def|get|set)(\s)(\[)/, ["keyword.$1", "white.entity", "$$", "@>def_dynamic_name/$1"]]
  ],
  func_: [
    [/export(?=\s+(get|set|def|global) )/, "keyword.export"],
    [/global(?=\s+(get|set|def) )/, "keyword.global"],
    [/(def)(\s)(@id)(\.)(@defid)/, [
      "keyword.$1",
      "white.entity",
      "identifier.target",
      "operator",
      "entity.name.def",
      "@>def_params&$1/$1"
    ]],
    [/(def)(\s)(@defid)/, ["keyword.$1", "white.entity", "entity.name.function.decl-const-func", "@>def_params&$1/$1"]]
  ],
  flow_: [
    [/(if|else|elif|unless)(?=\s|$)/, ["keyword.$1", "@flow_start=$1"]]
  ],
  flow_start: [
    denter({switchTo: "@>_flow&$F"}, -1, -1),
    [/[ \t]+/, "white"],
    "expr_"
  ],
  for_: [
    [/for(?: own)?@B/, "keyword.$#", "@for_start&forscope=decl-for"]
  ],
  while_: [
    [/(while|until)@B/, "keyword.$#", "@>while_body"]
  ],
  while_body: [
    denter(2, -1, 0),
    "block_"
  ],
  for_start: [
    denter({switchTo: "@>for_body"}, -1, -1),
    [/\[/, "array.[", "@array_var_body"],
    [/\{/, "object.{", "@object_body"],
    [/(@variable)/, "identifier.$F"],
    [/(\s*\,\s*)/, "separator", "@=decl-for-index"],
    [/\s(in|of)@B/, "keyword.$1", {switchTo: "@>for_source="}],
    [/[ \t]+/, "white"],
    "type_"
  ],
  for_source: [
    denter({switchTo: "@>for_body"}, -1, {switchTo: "@for_body"}),
    "expr_",
    [/[ \t]+/, "white"]
  ],
  for_body: [
    denter(2, -1, 0),
    "block_"
  ],
  decorator_: [
    [/(@decid)(\()/, ["decorator", "$2", "@_decorator_params"]],
    [/(@decid)/, "decorator"]
  ],
  _decorator_params: [
    [/\)/, ")", "@pop"],
    "params_"
  ],
  field_: [
    [/((?:lazy )?)((?:static )?)(const|let|attr)(?=\s|$)/, ["keyword.lazy", "keyword.static", "keyword.$1", "@_vardecl=field-$3"]],
    [/(static\s+)(?=@fieldid)/, "keyword.static"],
    [/(@fieldid)(?=$)/, "entity.name.field"],
    [/(@fieldid)/, ["entity.name.field", "@_field_1"]]
  ],
  _field_1: [
    denter(null, -1, -1),
    "type_",
    [/(\s*=)(?!\=)/, ["operator.assign", "@>_field_value&field"]],
    [/(\s*(?:\@)set\s*)/, ["keyword.spy", "@>_def&spy"]]
  ],
  _field_value: [
    denter(2, -1, 0),
    "block_",
    [/(\s*(?:\@)set\s*)/, ["@rematch", "@pop"]]
  ],
  var_: [
    [/((?:export )?)(const|let)(?=\s[\[\{\$a-zA-Z]|\s*$)/, ["keyword.export", "keyword.$1", "@_vardecl=decl-$2"]],
    [/((?:export )?)(const|let)(?=\s|$)/, ["keyword.export", "keyword.$1"]]
  ],
  inline_var_: [
    [/(const|let)(?=\s[\[\{\$a-zA-Z]|\s*$)/, ["keyword.$1", "@inline_var_body=decl-$1"]]
  ],
  string_: [
    [/"""/, "string", '@_herestring="""'],
    [/'''/, "string", "@_herestring='''"],
    [/["'`]/, "string.open", "@_string=$#"]
  ],
  number_: [
    [/0[xX][0-9a-fA-F_]+/, "number.hex"],
    [/0[b][01_]+/, "number.binary"],
    [/0[o][0-9_]+/, "number.octal"],
    [/(\d+)([a-z]+|\%)/, ["number", "unit"]],
    [/(\d*\.\d+(?:[eE][\-+]?\d+)?)([a-z]+|\%)/, ["number.float", "unit"]],
    [/\d+[eE]([\-+]?\d+)?/, "number.float"],
    [/\d[\d_]*\.\d[\d_]*([eE][\-+]?\d+)?/, "number.float"],
    [/\d[\d_]*/, "number.integer"],
    [/0[0-7]+(?!\d)/, "number.octal"],
    [/\d+/, "number"]
  ],
  _string: [
    [/[^"'\`\{\\]+/, "string"],
    [/@escapes/, "string.escape"],
    [/\./, "string.escape.invalid"],
    [/\{/, {cases: {
      "$F=='": "string",
      "@default": {token: "string.bracket.open", next: "@interpolation_body"}
    }}],
    [/["'`]/, {cases: {"$#==$F": {token: "string.close", next: "@pop"}, "@default": "string"}}],
    [/#/, "string"]
  ],
  _herestring: [
    [/("""|''')/, {cases: {"$1==$F": {token: "string", next: "@pop"}, "@default": "string"}}],
    [/[^#\\'"\{]+/, "string"],
    [/['"]+/, "string"],
    [/@escapes/, "string.escape"],
    [/\./, "string.escape.invalid"],
    [/\{/, {cases: {'$F=="""': {token: "string", next: "@interpolation_body"}, "@default": "string"}}],
    [/#/, "string"]
  ],
  interpolation_body: [
    [/\}/, "string.bracket.close", "@pop"],
    "expr_"
  ],
  _class: [
    denter(toodeep, -1, 0),
    "css_",
    "member_",
    "comment_",
    "decorator_",
    [/(get|set|def|static|prop|attr)@B/, "keyword.$0"],
    "field_",
    "common_"
  ],
  _tagclass: [
    "_class",
    [/(?=\<self)/, "entity.name.def.render", "@_render&def"]
  ],
  def_params: [
    [/\(/, "(", "@def_parens"],
    [/^/, "@rematch", {switchTo: "@_def"}],
    [/do@B/, "keyword.do", {switchTo: "@_def"}],
    "params_",
    [/@comment/, "comment"]
  ],
  def_parens: [
    [/\)/, ")", "@pop"],
    "params_"
  ],
  def_dynamic_name: [
    ["]", {token: "square.close", switchTo: "@def_params&$/"}],
    "expr_"
  ],
  _render: [
    denter(2, -1, -1),
    "block_"
  ],
  _def: [
    denter(2, -1, 0),
    "block_"
  ],
  _flow: [
    denter(2, -1, 0),
    "block_"
  ],
  _varblock: [
    denter(1, -1, -1),
    [/\[/, "array.[", "@array_var_body"],
    [/\{/, "object.{", "@object_body"],
    [/(@variable)/, "identifier.$F"],
    [/\s*\,\s*/, "separator"],
    [/(\s*\=\s*)(?=(for|while|until|if|unless|try)\s)/, "operator", "@pop"],
    [/(\s*\=\s*)/, "operator", "@var_value="],
    "type_",
    [/#(\s.*)?\n?$/, "comment"]
  ],
  _vardecl: [
    denter(null, -1, -1),
    [/\[/, "array.[", "@array_var_body"],
    [/\{/, "object.{", "@object_body"],
    [/(@variable)(?=\n|,|$)/, "identifier.$F", "@pop"],
    [/(@variable)/, "identifier.$F"],
    [/(\s*\=\s*)/, "operator.declval", {switchTo: "@var_value&value="}],
    "type_"
  ],
  array_var_body: [
    [/\]/, "]", "@pop"],
    [/\{/, "object.{", "@object_body"],
    [/\[/, "array.[", "@array_var_body"],
    "spread_",
    [/(@variable)/, "identifier.$F"],
    [/(\s*\=\s*)/, "operator.assign", "@array_var_body_value="],
    [",", "delimiter"]
  ],
  array_var_body_value: [
    [/(?=,|\)|]|})/, "delimiter", "@pop"],
    "expr_"
  ],
  inline_var_body: [
    [/\[/, "array.[", "@array_var_body"],
    [/\{/, "object.{", "@object_body"],
    [/(@variable)/, "identifier.$F"],
    [/(\s*\=\s*)/, "operator", "@pop"],
    "type_"
  ],
  var_value: [
    [/(?=,|\)|]|})/, "delimiter", "@pop"],
    denter({switchTo: "@>block"}, -1, -1),
    "block_"
  ],
  common_: [
    [/^(\t+)(?=\n|$)/, "white.tabs"],
    "whitespace"
  ],
  comma_: [
    [/\s*,\s*/, "delimiter.comma"]
  ],
  spread_: [
    [/\.\.\./, "operator.spread"]
  ],
  type_: [
    [/\\/, "@rematch", "@_type&-_type/0"]
  ],
  _type: [
    denter(-1, -1, -1),
    [/\\/, "delimiter.type.prefix"],
    [/\[/, "delimiter.type", "@/]"],
    [/\(/, "delimiter.type", "@/)"],
    [/\{/, "delimiter.type", "@/}"],
    [/\</, "delimiter.type", "@/>"],
    [/\|/, "delimiter.type.union"],
    [/\,|\s|\=|\./, {
      cases: {
        "$/==0": {token: "@rematch", next: "@pop"},
        "@default": "type"
      }
    }],
    [/[\]\}\)\>]/, {
      cases: {
        "$#==$/": {token: "delimiter.type", next: "@pop"},
        "@default": {token: "@rematch", next: "@pop"}
      }
    }],
    [/[\w\-\$]+/, "type"]
  ],
  css_: [
    [/global(?=\s+css@B)/, "keyword.$#"],
    [/css(?:\s+)?/, "keyword.css", "@>css_selector&rule-_sel"]
  ],
  sel_: [
    [/(\%)((?:@id)?)/, ["style.selector.mixin.prefix", "style.selector.mixin"]],
    [/(\@)(\.{0,2}[\w\-\<\>\!]*\+?)/, "style.selector.modifier"],
    [/(\@)(\.{0,2}[\w\-\<\>\!]*)/, "style.selector.modifier"],
    [/\.([\w\-]+)/, "style.selector.class-name"],
    [/\#([\w\-]+)/, "style.selector.id"],
    [/([\w\-]+)/, "style.selector.element"],
    [/(>+|~|\+)/, "style.selector.operator"],
    [/(\*+)/, "style.selector.element.any"],
    [/(\$)((?:@id)?)/, ["style.selector.reference.prefix", "style.selector.reference"]],
    [/\&/, "style.selector.context"],
    [/\(/, "delimiter.selector.parens.open", "@css_selector_parens"],
    [/\[/, "delimiter.selector.attr.open", "@css_selector_attr"],
    [/\s+/, "white"],
    [/,/, "style.selector.delimiter"],
    [/#(\s.*)?\n?$/, "comment"]
  ],
  css_props: [
    denter(null, -1, 0),
    [/(?=@cssPropertyKey2)/, "", "@css_property&-_styleprop-_stylepropkey"],
    [/#(\s.*)?\n?$/, "comment"],
    [/(?=[\%\*\w\&\$\>\.\[\@\!]|\#[\w\-])/, "", "@>css_selector&rule-_sel"],
    [/\s+/, "white"]
  ],
  css_selector: [
    denter({switchTo: "@css_props"}, -1, {token: "@rematch", switchTo: "@css_props&_props"}),
    [/(\}|\)|\])/, "@rematch", "@pop"],
    [/(?=\s*@cssPropertyKey2)/, "", {switchTo: "@css_props&_props"}],
    [/\s*#\s/, "@rematch", {switchTo: "@css_props&_props"}],
    "sel_"
  ],
  css_inline: [
    [/\]/, "style.close", "@pop"],
    [/(?=@cssPropertyKey2)/, "", "@css_property&-_styleprop-_stylepropkey"],
    [/(?=@cssPropertyPath\])/, "", "@css_property&-_styleprop-_stylepropkey"]
  ],
  css_selector_parens: [
    [/\)/, "delimiter.selector.parens.close", "@pop"],
    "sel_"
  ],
  css_selector_attr: [
    [/\]/, "delimiter.selector.parens.close", "@pop"],
    "sel_"
  ],
  css_property: [
    denter(null, -1, -1),
    [/\]/, "@rematch", "@pop"],
    [/(\d+)(@id)/, ["style.property.unit.number", "style.property.unit.name"]],
    [/((--|\$)@id)/, "style.property.var"],
    [/(-*@id)/, "style.property.name"],
    [/@cssModifier/, "style.property.modifier"],
    [/(\@+|\.+)(@id\-?)/, ["style.property.modifier.start", "style.property.modifier"]],
    [/\+(@id)/, "style.property.scope"],
    [/\s*([\:]\s*)(?=@br|$)/, "style.property.operator", {switchTo: "@>css_multiline_value&_stylevalue"}],
    [/\s*([\:]\s*)/, "style.property.operator", {switchTo: "@>css_value&_stylevalue"}]
  ],
  css_value_: [
    [/(x?xs|sm\-?|md\-?|lg\-?|xx*l|\dxl|hg|x+h)\b/, "style.value.size"],
    [/\#[0-9a-fA-F]+/, "style.value.color.hex"],
    [/((--|\$)@id)/, "style.value.var"],
    [/(@optid)(\@+|\.+)(@optid)/, ["style.property.name", "style.property.modifier.prefix", "style.property.modifier"]],
    "op_",
    "string_",
    "number_",
    "comment_",
    [/\s+/, "style.value.white"],
    [/\(/, "delimiter.style.parens.open", "@css_expressions"],
    [/\{/, "delimiter.style.curly.open", "@css_interpolation&-_styleinterpolation"],
    [/(@id)/, "style.value"]
  ],
  css_value: [
    denter({switchTo: "@>css_multiline_value"}, -1, -1),
    [/@cssPropertyKey2/, "@rematch", "@pop"],
    [/;/, "style.delimiter", "@pop"],
    [/(\}|\)|\])/, "@rematch", "@pop"],
    "css_value_"
  ],
  css_multiline_value: [
    denter(null, -1, 0),
    [/@cssPropertyKey2/, "invalid"],
    "css_value_"
  ],
  css_expressions: [
    [/\)/, "delimiter.style.parens.close", "@pop"],
    [/\(/, "delimiter.style.parens.open", "@css_expressions"],
    "css_value"
  ],
  css_interpolation: [
    [/\}/, "delimiter.style.curly.close", "@pop"],
    "expr_"
  ],
  expressions: [
    [/\,/, "delimiter.comma"]
  ],
  whitespace: [
    [/[\r\n]+/, "br"],
    [/[ \t\r\n]+/, "white"]
  ],
  space: [
    [/[ \t]+/, "white"]
  ],
  tag_: [
    [/(\s*)(<)(?=\.)/, ["white", "tag.open", "@_tag/flag"]],
    [/(\s*)(<)(?=\w|\{|\[|\%|\#|>)/, ["white", "tag.open", "@_tag/name"]]
  ],
  tag_content: [
    denter(2, -1, 0),
    [/\)|\}|\]/, "@rematch", "@pop"],
    "common_",
    "flow_",
    "var_",
    "for_",
    "css_",
    "expr_",
    "do_"
  ],
  tag_children: [],
  _tag: [
    [/\/>/, "tag.close", "@pop"],
    [/>/, "tag.close", {switchTo: "@>tag_content=&-_tagcontent"}],
    [/>/, "tag.close", "@pop"],
    [/(\-?\d+)/, "tag.$S3"],
    [/(\%)(@id)/, ["tag.mixin.prefix", "tag.mixin"]],
    [/\#@id/, "tag.id"],
    [/\./, {cases: {
      "$/==event": {token: "tag.event-modifier.start", switchTo: "@/event-modifier"},
      "$/==event-modifier": {token: "tag.event-modifier.start", switchTo: "@/event-modifier"},
      "$/==modifier": {token: "tag.modifier.start", switchTo: "@/modifier"},
      "$/==rule": {token: "tag.rule-modifier.start", switchTo: "@/rule-modifier"},
      "$/==rule-modifier": {token: "tag.rule-modifier.start", switchTo: "@/rule-modifier"},
      "@default": {token: "tag.flag.start", switchTo: "@/flag"}
    }}],
    [/(\$@id)/, {cases: {
      "$/==name": "tag.reference",
      "@default": "tag.$/"
    }}],
    [/\{/, "tag.$/.interpolation.open", "@_tag_interpolation"],
    [/\[/, "style.open", "@css_inline"],
    [/(\s*\=\s*)/, "operator.equals.tagop.tag-$/", "@_tag_value&-value"],
    [/\:/, {token: "tag.event.start", switchTo: "@/event"}],
    "tag_event_",
    [/(\-?@tagIdentifier)(\:@id)?/, {cases: {
      "$/==attr": {token: "@rematch", next: "@_tag_attr&-_tagattr"},
      "@default": {token: "tag.$/"}
    }}],
    [/\(/, {token: "tag.$/.parens.open", next: "@_tag_parens/0"}],
    [/\s+/, {token: "tag.white", switchTo: "@/attr"}],
    "comment_"
  ],
  tag_event_: [
    [/(?=\@@optid)/, "", "@_tag_event&-_listener"]
  ],
  _tag_part: [
    [/\)|\}|\]|\>/, "@rematch", "@pop"]
  ],
  _tag_event: [
    "_tag_part",
    [/(\@)(@optid)/, ["tag.event.start", "tag.event.name"]],
    [/(\.)(@optid)/, ["tag.event-modifier.start", "tag.event-modifier.name"]],
    [/\(/, {token: "tag.$/.parens.open", next: "@_tag_parens/0"}],
    [/(\s*\=\s*)/, "operator.equals.tagop.tag-$/", "@_tag_value&handler"],
    [/\s+/, "@rematch", "@pop"]
  ],
  tag_attr_: [
    [/(?=@tagIdentifier(\:@id)?)/, "", "@_tag_attr&-_attribute"]
  ],
  _tag_attr: [
    "_tag_part",
    [/(\-?@tagIdentifier)(\:@id)?/, "tag.attr"],
    [/\.(@optid)/, "tag.event-modifierzz"],
    [/\(/, {token: "tag.parens.open.$/", next: "@_tag_parens/0"}],
    [/(\s*\=\s*)/, "operator.equals.tagop.tag-$/", "@_tag_value&-tagattrvalue"],
    [/\s+/, "@rematch", "@pop"]
  ],
  _tag_interpolation: [
    [/\}/, "tag.$/.interpolation.close", "@pop"],
    "expr_",
    [/\)|\]/, "invalid"]
  ],
  _tag_parens: [
    [/\)/, "tag.$/.parens.close", "@pop"],
    "arglist_",
    [/\]|\}/, "invalid"]
  ],
  _tag_value: [
    [/(?=(\/?\>|\s))/, "", "@pop"],
    "attr_expr_"
  ],
  regexp_: [
    [/\/(?!\ )(?=([^\\\/]|\\.)+\/)/, {token: "regexp.slash.open", bracket: "@open", next: "@_regexp"}],
    [/\/\/\//, {token: "regexp.slash.open", bracket: "@open", next: "@_hereregexp"}],
    [/(\/)(\/)/, ["regexp.slash.open", "regexp.slash.close"]]
  ],
  _regexp: [
    [/(\{)(\d+(?:,\d*)?)(\})/, ["regexp.escape.control", "regexp.escape.control", "regexp.escape.control"]],
    [/(\[)(\^?)(?=(?:[^\]\\\/]|\\.)+)/, ["regexp.escape.control", {token: "regexp.escape.control", next: "@_regexrange"}]],
    [/(\()(\?:|\?=|\?!)/, ["regexp.escape.control", "regexp.escape.control"]],
    [/[()]/, "regexp.escape.control"],
    [/@regexpctl/, "regexp.escape.control"],
    [/[^\\\/]/, "regexp"],
    [/@regexpesc/, "regexp.escape"],
    [/\\:/, "regexp.escape"],
    [/\\\./, "regexp.invalid"],
    [/(\/)(\w+)/, [{token: "regexp.slash.close"}, {token: "regexp.flags", next: "@pop"}]],
    ["/", {token: "regexp.slash.close", next: "@pop"}],
    [/./, "regexp.invalid"]
  ],
  _regexrange: [
    [/-/, "regexp.escape.control"],
    [/\^/, "regexp.invalid"],
    [/@regexpesc/, "regexp.escape"],
    [/[^\]]/, "regexp"],
    [/\]/, "regexp.escape.control", "@pop"]
  ],
  _hereregexp: [
    [/[^\\\/#]/, "regexp"],
    [/\\./, "regexp"],
    [/#.*$/, "comment"],
    ["///[igm]*", "regexp", "@pop"],
    [/\//, "regexp"],
    "comment_"
  ]
};
function rewriteΞstate(raw) {
  let state = ["$S1", "$S2", "$S3", "$S4", "$S5", "$S6"];
  if (raw.match(/\@(pop|push|popall)/)) {
    return raw;
  }
  ;
  if (raw[0] == "@") {
    raw = raw.slice(1);
  }
  ;
  if (raw.indexOf(".") >= 0) {
    return raw;
  }
  ;
  raw = rewriteΞtoken(raw);
  if (raw[0] == ">") {
    state[1] = "$S6	";
    raw = raw.slice(1);
  }
  ;
  for (let φ52 = 0, φ6 = iter$__2(raw.split(/(?=[\/\&\=\*])/)), φ72 = φ6.length; φ52 < φ72; φ52++) {
    let part = φ6[φ52];
    if (part[0] == "&") {
      if (part[1] == "-" || part[1] == "_") {
        state[2] = "$S3" + part.slice(1);
      } else {
        state[2] = "$S3-" + part.slice(1);
      }
      ;
    } else if (part[0] == "+") {
      state[3] = "$S4-" + part.slice(1);
    } else if (part[0] == "=") {
      state[3] = part.slice(1);
    } else if (part[0] == "/") {
      state[4] = part.slice(1);
    } else if (part[0] == "*") {
      state[5] = part.slice(1);
    } else {
      state[0] = part;
    }
    ;
  }
  ;
  return state.join(".");
}
function rewriteΞtoken(raw) {
  let orig = raw;
  raw = raw.replace("$/", "$S5");
  raw = raw.replace("$F", "$S4");
  raw = raw.replace("$&", "$S3");
  raw = raw.replace("$I", "$S2");
  raw = raw.replace("$T", "$S2");
  return raw;
}
function rewriteΞactions(actions, add) {
  if (typeof actions == "string") {
    actions = {token: actions};
  }
  ;
  if (actions && actions.token != void 0) {
    actions.token = rewriteΞtoken(actions.token);
    if (typeof add == "string") {
      actions.next = add;
    } else if (add) {
      Object.assign(actions, add);
    }
    ;
    if (actions.next) {
      actions.next = rewriteΞstate(actions.next);
    }
    ;
    if (actions.switchTo) {
      actions.switchTo = rewriteΞstate(actions.switchTo);
    }
    ;
  } else if (actions && actions.cases) {
    let cases = {};
    for (let φ102 = actions.cases, φ82 = 0, φ92 = Object.keys(φ102), φ11 = φ92.length, k, v; φ82 < φ11; φ82++) {
      k = φ92[φ82];
      v = φ102[k];
      let newkey = rewriteΞtoken(k);
      cases[newkey] = rewriteΞactions(v);
    }
    ;
    actions.cases = cases;
  } else if (actions instanceof Array) {
    let result = [];
    let curr = null;
    for (let i = 0, φ122 = iter$__2(actions), φ132 = φ122.length; i < φ132; i++) {
      let action = φ122[i];
      if (action[0] == "@" && i == actions.length - 1 && curr) {
        action = {next: action};
      }
      ;
      if (typeof action == "object") {
        if (action.token != void 0 || action.cases) {
          result.push(curr = Object.assign({}, action));
        } else {
          Object.assign(curr, action);
        }
        ;
      } else if (typeof action == "string") {
        result.push(curr = {token: rewriteΞtoken(action)});
      }
      ;
    }
    ;
    actions = result;
  }
  ;
  if (actions instanceof Array) {
    for (let i = 0, φ142 = iter$__2(actions), φ152 = φ142.length; i < φ152; i++) {
      let action = φ142[i];
      if (action.token && action.token.indexOf("$$") >= 0) {
        action.token = action.token.replace("$$", "$" + (i + 1));
      }
      ;
      if (action.next) {
        action.next = rewriteΞstate(action.next);
      }
      ;
      if (action.switchTo) {
        action.switchTo = rewriteΞstate(action.switchTo);
      }
      ;
    }
    ;
  }
  ;
  return actions;
}
for (let φ16 = 0, φ17 = Object.keys(states), φ18 = φ17.length, key, rules; φ16 < φ18; φ16++) {
  key = φ17[φ16];
  rules = states[key];
  let i = 0;
  while (i < rules.length) {
    let rule = rules[i];
    if (rule[0] instanceof Array) {
      rules.splice(i, 1, ...rule);
      continue;
    } else if (typeof rule == "string") {
      rules[i] = {include: rule};
    } else if (rule[1] instanceof Array) {
      rule[1] = rewriteΞactions(rule[1]);
    } else if (rule instanceof Array) {
      rule.splice(1, 2, rewriteΞactions(rule[1], rule[2]));
    }
    ;
    i++;
  }
  ;
}
var grammar = {
  defaultToken: "invalid",
  ignoreCase: false,
  tokenPostfix: "",
  brackets: [
    {open: "{", close: "}", token: "bracket.curly"},
    {open: "[", close: "]", token: "bracket.square"},
    {open: "(", close: ")", token: "bracket.parenthesis"}
  ],
  keywords: [
    "def",
    "and",
    "or",
    "is",
    "isnt",
    "not",
    "on",
    "yes",
    "@",
    "no",
    "off",
    "true",
    "false",
    "null",
    "this",
    "self",
    "as",
    "new",
    "delete",
    "typeof",
    "in",
    "instanceof",
    "return",
    "throw",
    "break",
    "continue",
    "debugger",
    "if",
    "elif",
    "else",
    "switch",
    "for",
    "while",
    "do",
    "try",
    "catch",
    "finally",
    "class",
    "extends",
    "super",
    "undefined",
    "then",
    "unless",
    "until",
    "loop",
    "of",
    "by",
    "when",
    "tag",
    "prop",
    "attr",
    "export",
    "import",
    "extend",
    "var",
    "let",
    "const",
    "require",
    "isa",
    "await"
  ],
  boolean: ["true", "false", "yes", "no", "undefined", "null"],
  operators: [
    "=",
    "!",
    "~",
    "?",
    ":",
    "!!",
    "??",
    "&",
    "|",
    "^",
    "%",
    "<<",
    "!&",
    ">>",
    ">>>",
    "+=",
    "-=",
    "*=",
    "/=",
    "&=",
    "|=",
    "?=",
    "??=",
    "^=",
    "%=",
    "~=",
    "<<=",
    ">>=",
    ">>>=",
    "..",
    "...",
    "||=",
    `&&=`,
    "**=",
    "**",
    "|=?",
    "~=?",
    "^=?",
    "=?",
    "and",
    "or"
  ],
  assignments: [
    "=",
    "|=?",
    "~=?",
    "^=?",
    "=?",
    "^=",
    "%=",
    "~=",
    "<<=",
    ">>=",
    ">>>=",
    "||=",
    `&&=`,
    "?=",
    "??=",
    "+=",
    "-=",
    "*=",
    "/=",
    "&=",
    "|=",
    "**="
  ],
  logic: [
    ">",
    "<",
    "==",
    "<=",
    ">=",
    "!=",
    "&&",
    "||",
    "===",
    "!=="
  ],
  ranges: ["..", "..."],
  spread: ["..."],
  dot: ["."],
  access: [".", ".."],
  math: ["+", "-", "*", "/", "++", "--"],
  unspaced_ops: regexify("... . .. + * ++ --"),
  comment: /#(\s.*)?(\n|$)/,
  symbols: /[=><!~?&%|+\-*\^,]+/,
  escapes: /\\(?:[abfnrtv\\"'$]|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,
  postaccess: /(:(?=\w))?/,
  ivar: /\@[a-zA-Z_]\w*/,
  B: /(?=\s|$)/,
  br: /[\r\n]+/,
  constant: /[A-Z][\w\$]*(?:\-+[\w\$]+)*/,
  id: /[A-Za-z_\$][\w\$]*(?:\-+[\w\$]+)*\??/,
  plainid: /[A-Za-z_\$][\w\$]*(?:\-+[\w\$]+)*\??/,
  fieldid: /[\@\#]*@plainid/,
  propid: /[\@\#]*@plainid/,
  defid: /[\@\#]*@plainid/,
  decid: /\@@plainid/,
  symid: /\#+@plainid/,
  symref: /\#\#@plainid/,
  optid: /(?:@id)?/,
  esmIdentifier: /[A-Za-z_\$\@][\w\$]*(?:\-+[\w\$]+)*\??/,
  propertyPath: /(?:[A-Za-z_\$][A-Za-z\d\-\_\$]*\.)?(?:[A-Za-z_\$][A-Za-z\d\-\_\$]*)/,
  tagNameIdentifier: /(?:[\w\-]+\:)?\w+(?:\-\w+)*/,
  variable: /[\w\$]+(?:-[\w\$]*)*\??/,
  varKeyword: /var|let|const/,
  tagIdentifier: /-*[a-zA-Z][\w\-]*/,
  implicitCall: /(?!\s(?:and|or|is|isa)\s)(?=\s[\w\'\"\/\[\{])/,
  cssModifier: /(?:\@+[\<\>\!]?[\w\-]+\+?|\.+@id\-?)/,
  cssPropertyPath: /[\@\.]*[\w\-\$]+(?:[\@\.]+[\w\-\$]+)*/,
  cssPropertyKey: /[\@\.]*[\w\-\$]+(?:[\@\.]+[\w\-\$]+)*(?:\s*\:)/,
  cssVariable: /(?:--|\$)[\w\-\$]+/,
  cssPropertyName: /[\w\-\$]+/,
  cssPropertyKey2: /(?:@cssPropertyName(?:@cssModifier)*|@cssModifier+)(?:\s*\:)/,
  cssUpModifier: /\.\.[\w\-\$]+/,
  cssIsModifier: /\.[\w\-\$]+/,
  regEx: /\/(?!\/\/)(?:[^\/\\]|\\.)*\/[igm]*/,
  regexpctl: /[(){}\[\]\$\^|\-*+?\.]/,
  regexpesc: /\\(?:[bBdDfnrstvwWn0\\\/]|@regexpctl|c[A-Z]|x[0-9a-fA-F]{2}|u[0-9a-fA-F]{4})/,
  tokenizer: states
};

// src/program/monarch/common.ts
var MonarchBracket;
(function(MonarchBracket2) {
  MonarchBracket2[MonarchBracket2["None"] = 0] = "None";
  MonarchBracket2[MonarchBracket2["Open"] = 1] = "Open";
  MonarchBracket2[MonarchBracket2["Close"] = -1] = "Close";
})(MonarchBracket || (MonarchBracket = {}));
function isFuzzyActionArr(what) {
  return Array.isArray(what);
}
function isFuzzyAction(what) {
  return !isFuzzyActionArr(what);
}
function isString(what) {
  return typeof what === "string";
}
function isIAction(what) {
  return !isString(what);
}
function empty(s) {
  return s ? false : true;
}
function fixCase(lexer2, str) {
  return lexer2.ignoreCase && str ? str.toLowerCase() : str;
}
function sanitize(s) {
  return s.replace(/[&<>'"_]/g, "-");
}
function log(lexer2, msg) {
  console.log(`${lexer2.languageId}: ${msg}`);
}
function createError(lexer2, msg) {
  return new Error(`${lexer2.languageId}: ${msg}`);
}
var substitutionCache = {};
function compileSubstitution(str) {
  const parts = [];
  let i = 0;
  let l = str.length;
  let part = "";
  let sub = 0;
  while (i < l) {
    let chr = str[i++];
    if (chr == "$") {
      let next = str[i++];
      if (next == "$") {
        part += "$";
        continue;
      }
      if (part)
        parts.push(part);
      part = "";
      if (next == "#") {
        parts.push(0);
      } else if (next == "S") {
        parts.push(parseInt(str[i++]) + 100);
      } else {
        parts.push(parseInt(next) + 1);
      }
    } else {
      part += chr;
    }
  }
  if (part)
    parts.push(part);
  substitutionCache[str] = parts;
  return parts;
}
function substituteMatches(lexer2, str, id, matches, state) {
  let stateMatches = null;
  let parts = substitutionCache[str] || compileSubstitution(str);
  let out = "";
  for (let i = 0; i < parts.length; i++) {
    let part = parts[i];
    if (typeof part == "string") {
      out += part;
    } else if (part > 100) {
      if (stateMatches === null)
        stateMatches = state.split(".");
      out += stateMatches[part - 101] || "";
    } else if (part === 100) {
      out += state;
    } else if (part === 0) {
      out += id;
    } else if (part > 0) {
      out += matches[part - 1];
    }
  }
  return out;
}
var FIND_RULES_MAP = {};
function findRules(lexer2, inState) {
  let state = inState;
  if (FIND_RULES_MAP[state]) {
    return lexer2.tokenizer[FIND_RULES_MAP[state]];
  }
  while (state && state.length > 0) {
    const rules = lexer2.tokenizer[state];
    if (rules) {
      FIND_RULES_MAP[inState] = state;
      return rules;
    }
    const idx = state.lastIndexOf(".");
    if (idx < 0) {
      state = null;
    } else {
      state = state.substr(0, idx);
    }
  }
  return null;
}
function stateExists(lexer2, inState) {
  let state = inState;
  while (state && state.length > 0) {
    const exist = lexer2.stateNames[state];
    if (exist) {
      return true;
    }
    const idx = state.lastIndexOf(".");
    if (idx < 0) {
      state = null;
    } else {
      state = state.substr(0, idx);
    }
  }
  return false;
}

// src/program/monarch/compile.ts
function isArrayOf(elemType, obj) {
  if (!obj) {
    return false;
  }
  if (!Array.isArray(obj)) {
    return false;
  }
  for (const el of obj) {
    if (!elemType(el)) {
      return false;
    }
  }
  return true;
}
function bool(prop, defValue) {
  if (typeof prop === "boolean") {
    return prop;
  }
  return defValue;
}
function string(prop, defValue) {
  if (typeof prop === "string") {
    return prop;
  }
  return defValue;
}
function arrayToHash(array) {
  const result = {};
  for (const e of array) {
    result[e] = true;
  }
  return result;
}
function createKeywordMatcher(arr, caseInsensitive = false) {
  if (caseInsensitive) {
    arr = arr.map(function(x) {
      return x.toLowerCase();
    });
  }
  const hash = arrayToHash(arr);
  if (caseInsensitive) {
    return function(word) {
      return hash[word.toLowerCase()] !== void 0 && hash.hasOwnProperty(word.toLowerCase());
    };
  } else {
    return function(word) {
      return hash[word] !== void 0 && hash.hasOwnProperty(word);
    };
  }
}
function compileRegExp(lexer2, str) {
  let n = 0;
  while (str.indexOf("@") >= 0 && n < 5) {
    n++;
    str = str.replace(/@(\w+)/g, function(s, attr) {
      let sub = "";
      if (typeof lexer2[attr] === "string") {
        sub = lexer2[attr];
      } else if (lexer2[attr] && lexer2[attr] instanceof RegExp) {
        sub = lexer2[attr].source;
      } else {
        if (lexer2[attr] === void 0) {
          throw createError(lexer2, "language definition does not contain attribute '" + attr + "', used at: " + str);
        } else {
          throw createError(lexer2, "attribute reference '" + attr + "' must be a string, used at: " + str);
        }
      }
      return empty(sub) ? "" : "(?:" + sub + ")";
    });
  }
  return new RegExp(str, lexer2.ignoreCase ? "i" : "");
}
function selectScrutinee(id, matches, state, num) {
  if (num < 0) {
    return id;
  }
  if (num < matches.length) {
    return matches[num];
  }
  if (num >= 100) {
    num = num - 100;
    let parts = state.split(".");
    parts.unshift(state);
    if (num < parts.length) {
      return parts[num];
    }
  }
  return null;
}
function createGuard(lexer2, ruleName, tkey, val) {
  let scrut = -1;
  let oppat = tkey;
  let matches = tkey.match(/^\$(([sS]?)(\d\d?)|#)(.*)$/);
  if (matches) {
    if (matches[3]) {
      scrut = parseInt(matches[3]);
      if (matches[2]) {
        scrut = scrut + 100;
      }
    }
    oppat = matches[4];
  }
  let op = "~";
  let pat = oppat;
  if (!oppat || oppat.length === 0) {
    op = "!=";
    pat = "";
  } else if (/^\w*$/.test(pat)) {
    op = "==";
  } else {
    matches = oppat.match(/^(@|!@|~|!~|==|!=)(.*)$/);
    if (matches) {
      op = matches[1];
      pat = matches[2];
    }
  }
  let tester;
  if ((op === "~" || op === "!~") && /^(\w|\|)*$/.test(pat)) {
    let inWords = createKeywordMatcher(pat.split("|"), lexer2.ignoreCase);
    tester = function(s) {
      return op === "~" ? inWords(s) : !inWords(s);
    };
  } else if (op === "@" || op === "!@") {
    let words = lexer2[pat];
    if (!words) {
      throw createError(lexer2, "the @ match target '" + pat + "' is not defined, in rule: " + ruleName);
    }
    if (!isArrayOf(function(elem) {
      return typeof elem === "string";
    }, words)) {
      throw createError(lexer2, "the @ match target '" + pat + "' must be an array of strings, in rule: " + ruleName);
    }
    let inWords = createKeywordMatcher(words, lexer2.ignoreCase);
    tester = function(s) {
      return op === "@" ? inWords(s) : !inWords(s);
    };
  } else if (op === "~" || op === "!~") {
    if (pat.indexOf("$") < 0) {
      let re = compileRegExp(lexer2, "^" + pat + "$");
      tester = function(s) {
        return op === "~" ? re.test(s) : !re.test(s);
      };
    } else {
      tester = function(s, id, matches2, state) {
        let re = compileRegExp(lexer2, "^" + substituteMatches(lexer2, pat, id, matches2, state) + "$");
        return re.test(s);
      };
    }
  } else {
    if (pat.indexOf("$") < 0) {
      let patx = fixCase(lexer2, pat);
      tester = function(s) {
        return op === "==" ? s === patx : s !== patx;
      };
    } else {
      let patx = fixCase(lexer2, pat);
      tester = function(s, id, matches2, state, eos) {
        let patexp = substituteMatches(lexer2, patx, id, matches2, state);
        return op === "==" ? s === patexp : s !== patexp;
      };
    }
  }
  if (scrut === -1) {
    return {
      name: tkey,
      value: val,
      test: function(id, matches2, state, eos) {
        return tester(id, id, matches2, state, eos);
      }
    };
  } else {
    return {
      name: tkey,
      value: val,
      test: function(id, matches2, state, eos) {
        let scrutinee = selectScrutinee(id, matches2, state, scrut);
        return tester(!scrutinee ? "" : scrutinee, id, matches2, state, eos);
      }
    };
  }
}
function compileAction(lexer2, ruleName, action) {
  if (!action) {
    return {token: ""};
  } else if (typeof action === "string") {
    return action;
  } else if (action.token || action.token === "") {
    if (typeof action.token !== "string") {
      throw createError(lexer2, "a 'token' attribute must be of type string, in rule: " + ruleName);
    } else {
      let newAction = {token: action.token};
      if (action.token.indexOf("$") >= 0) {
        newAction.tokenSubst = true;
      }
      if (typeof action.bracket === "string") {
        if (action.bracket === "@open") {
          newAction.bracket = MonarchBracket.Open;
        } else if (action.bracket === "@close") {
          newAction.bracket = MonarchBracket.Close;
        } else {
          throw createError(lexer2, "a 'bracket' attribute must be either '@open' or '@close', in rule: " + ruleName);
        }
      }
      if (action.next) {
        if (typeof action.next !== "string") {
          throw createError(lexer2, "the next state must be a string value in rule: " + ruleName);
        } else {
          let next = action.next;
          if (!/^(@pop|@push|@popall)$/.test(next)) {
            if (next[0] === "@") {
              next = next.substr(1);
            }
            if (next.indexOf("$") < 0) {
              if (!stateExists(lexer2, substituteMatches(lexer2, next, "", [], ""))) {
                throw createError(lexer2, "the next state '" + action.next + "' is not defined in rule: " + ruleName);
              }
            }
          }
          newAction.next = next;
        }
      }
      if (typeof action.goBack === "number") {
        newAction.goBack = action.goBack;
      }
      if (typeof action.switchTo === "string") {
        newAction.switchTo = action.switchTo;
      }
      if (typeof action.log === "string") {
        newAction.log = action.log;
      }
      if (typeof action._push === "string") {
        newAction._push = action._push;
      }
      if (typeof action._pop === "string") {
        newAction._pop = action._pop;
      }
      if (typeof action.mark === "string") {
        newAction.mark = action.mark;
      }
      if (typeof action.fn === "string") {
        newAction.fn = action.fn;
      }
      if (typeof action.nextEmbedded === "string") {
        newAction.nextEmbedded = action.nextEmbedded;
        lexer2.usesEmbedded = true;
      }
      return newAction;
    }
  } else if (Array.isArray(action)) {
    let results = [];
    for (let i = 0, len = action.length; i < len; i++) {
      results[i] = compileAction(lexer2, ruleName, action[i]);
    }
    return {group: results};
  } else if (action.cases) {
    let cases = [];
    for (let tkey in action.cases) {
      if (action.cases.hasOwnProperty(tkey)) {
        const val = compileAction(lexer2, ruleName, action.cases[tkey]);
        if (tkey === "@default" || tkey === "@" || tkey === "") {
          cases.push({test: void 0, value: val, name: tkey});
        } else if (tkey === "@eos") {
          cases.push({test: function(id, matches, state, eos) {
            return eos;
          }, value: val, name: tkey});
        } else {
          cases.push(createGuard(lexer2, ruleName, tkey, val));
        }
      }
    }
    const def = lexer2.defaultToken;
    return {
      test: function(id, matches, state, eos) {
        for (const _case of cases) {
          const didmatch = !_case.test || _case.test(id, matches, state, eos);
          if (didmatch) {
            return _case.value;
          }
        }
        return def;
      }
    };
  } else {
    throw createError(lexer2, "an action must be a string, an object with a 'token' or 'cases' attribute, or an array of actions; in rule: " + ruleName);
  }
}
var Rule = class {
  constructor(name) {
    this.regex = new RegExp("");
    this.action = {token: ""};
    this.matchOnlyAtLineStart = false;
    this.name = "";
    this.name = name;
    this.stats = {time: 0, count: 0, hits: 0};
  }
  setRegex(lexer2, re) {
    let sregex;
    if (typeof re === "string") {
      sregex = re;
    } else if (re instanceof RegExp) {
      sregex = re.source;
    } else {
      throw createError(lexer2, "rules must start with a match string or regular expression: " + this.name);
    }
    if (sregex.length == 2 && sregex[0] == "\\" && /[\{\}\(\)\[\]]/.test(sregex[1])) {
      this.string = sregex[1];
    }
    this.matchOnlyAtLineStart = sregex.length > 0 && sregex[0] === "^";
    this.name = this.name + ": " + sregex;
    this.regex = compileRegExp(lexer2, "^(?:" + (this.matchOnlyAtLineStart ? sregex.substr(1) : sregex) + ")");
  }
  setAction(lexer2, act) {
    this.action = compileAction(lexer2, this.name, act);
  }
};
function compile(languageId, json) {
  if (!json || typeof json !== "object") {
    throw new Error("Monarch: expecting a language definition object");
  }
  let lexer2 = {};
  lexer2.languageId = languageId;
  lexer2.noThrow = false;
  lexer2.maxStack = 100;
  lexer2.start = typeof json.start === "string" ? json.start : null;
  lexer2.ignoreCase = bool(json.ignoreCase, false);
  lexer2.tokenPostfix = string(json.tokenPostfix, "." + lexer2.languageId);
  lexer2.defaultToken = string(json.defaultToken, "source");
  lexer2.usesEmbedded = false;
  let lexerMin = json;
  lexerMin.languageId = languageId;
  lexerMin.ignoreCase = lexer2.ignoreCase;
  lexerMin.noThrow = lexer2.noThrow;
  lexerMin.usesEmbedded = lexer2.usesEmbedded;
  lexerMin.stateNames = json.tokenizer;
  lexerMin.defaultToken = lexer2.defaultToken;
  function addRules(state, newrules, rules) {
    for (const rule of rules) {
      let include = rule.include;
      if (include) {
        if (typeof include !== "string") {
          throw createError(lexer2, "an 'include' attribute must be a string at: " + state);
        }
        if (include[0] === "@") {
          include = include.substr(1);
        }
        if (!json.tokenizer[include]) {
          throw createError(lexer2, "include target '" + include + "' is not defined at: " + state);
        }
        addRules(state + "." + include, newrules, json.tokenizer[include]);
      } else {
        const newrule = new Rule(state);
        if (Array.isArray(rule) && rule.length >= 1 && rule.length <= 3) {
          newrule.setRegex(lexerMin, rule[0]);
          if (rule.length >= 3) {
            if (typeof rule[1] === "string") {
              newrule.setAction(lexerMin, {token: rule[1], next: rule[2]});
            } else if (typeof rule[1] === "object") {
              const rule1 = rule[1];
              rule1.next = rule[2];
              newrule.setAction(lexerMin, rule1);
            } else {
              throw createError(lexer2, "a next state as the last element of a rule can only be given if the action is either an object or a string, at: " + state);
            }
          } else {
            newrule.setAction(lexerMin, rule[1]);
          }
        } else {
          if (!rule.regex) {
            throw createError(lexer2, "a rule must either be an array, or an object with a 'regex' or 'include' field at: " + state);
          }
          if (rule.name) {
            if (typeof rule.name === "string") {
              newrule.name = rule.name;
            }
          }
          if (rule.matchOnlyAtStart) {
            newrule.matchOnlyAtLineStart = bool(rule.matchOnlyAtLineStart, false);
          }
          newrule.setRegex(lexerMin, rule.regex);
          newrule.setAction(lexerMin, rule.action);
        }
        newrules.push(newrule);
      }
    }
  }
  if (!json.tokenizer || typeof json.tokenizer !== "object") {
    throw createError(lexer2, "a language definition must define the 'tokenizer' attribute as an object");
  }
  lexer2.tokenizer = [];
  for (let key in json.tokenizer) {
    if (json.tokenizer.hasOwnProperty(key)) {
      if (!lexer2.start) {
        lexer2.start = key;
      }
      const rules = json.tokenizer[key];
      lexer2.tokenizer[key] = new Array();
      addRules("tokenizer." + key, lexer2.tokenizer[key], rules);
    }
  }
  lexer2.usesEmbedded = lexerMin.usesEmbedded;
  if (json.brackets) {
    if (!Array.isArray(json.brackets)) {
      throw createError(lexer2, "the 'brackets' attribute must be defined as an array");
    }
  } else {
    json.brackets = [
      {open: "{", close: "}", token: "delimiter.curly"},
      {open: "[", close: "]", token: "delimiter.square"},
      {open: "(", close: ")", token: "delimiter.parenthesis"},
      {open: "<", close: ">", token: "delimiter.angle"}
    ];
  }
  let brackets = [];
  for (let el of json.brackets) {
    let desc = el;
    if (desc && Array.isArray(desc) && desc.length === 3) {
      desc = {token: desc[2], open: desc[0], close: desc[1]};
    }
    if (desc.open === desc.close) {
      throw createError(lexer2, "open and close brackets in a 'brackets' attribute must be different: " + desc.open + "\n hint: use the 'bracket' attribute if matching on equal brackets is required.");
    }
    if (typeof desc.open === "string" && typeof desc.token === "string" && typeof desc.close === "string") {
      brackets.push({
        token: desc.token + lexer2.tokenPostfix,
        open: fixCase(lexer2, desc.open),
        close: fixCase(lexer2, desc.close)
      });
    } else {
      throw createError(lexer2, "every element in the 'brackets' array must be a '{open,close,token}' object or array");
    }
  }
  lexer2.brackets = brackets;
  lexer2.noThrow = true;
  return lexer2;
}

// src/program/monarch/token.ts
var Token = class {
  constructor(offset, type, language) {
    this.offset = offset | 0;
    this.type = type;
    this.language = language;
    this.kind = 0;
    this.mods = 0;
    this.value = null;
    this.stack = null;
  }
  toString() {
    return this.value || "";
  }
  get span() {
    return {offset: this.offset, length: this.value ? this.value.length : 0};
  }
  get indent() {
    return 0;
  }
  get startOffset() {
    return this.offset;
  }
  get endOffset() {
    return this.offset + (this.value ? this.value.length : 0);
  }
  clone() {
    let tok = new Token(this.offset, this.type, this.language);
    tok.value = this.value;
    tok.stack = this.stack;
    return tok;
  }
  match(val) {
    if (typeof val == "string") {
      if (val.indexOf(" ") > 0) {
        val = val.split(" ");
      } else {
        let idx = this.type.indexOf(val);
        return val[0] == "." ? idx >= 0 : idx == 0;
      }
    }
    if (val instanceof Array) {
      for (let item of val) {
        let idx = this.type.indexOf(item);
        let hit = item[0] == "." ? idx >= 0 : idx == 0;
        if (hit)
          return true;
      }
    }
    if (val instanceof RegExp) {
      return val.test(this.type);
    }
    return false;
  }
};
var TokenizationResult = class {
  constructor(tokens, endState) {
    this.tokens = tokens;
    this.endState = endState;
  }
};

// src/program/monarch/lexer.ts
var CACHE_STACK_DEPTH = 10;
function statePart(state, index) {
  return state.split(".")[index];
}
var _MonarchStackElementFactory = class {
  static create(parent, state) {
    return this._INSTANCE.create(parent, state);
  }
  constructor(maxCacheDepth) {
    this._maxCacheDepth = maxCacheDepth;
    this._entries = Object.create(null);
  }
  create(parent, state) {
    if (parent !== null && parent.depth >= this._maxCacheDepth) {
      return new MonarchStackElement(parent, state);
    }
    let stackElementId = MonarchStackElement.getStackElementId(parent);
    if (stackElementId.length > 0) {
      stackElementId += "|";
    }
    stackElementId += state;
    let result = this._entries[stackElementId];
    if (result) {
      return result;
    }
    result = new MonarchStackElement(parent, state);
    this._entries[stackElementId] = result;
    return result;
  }
};
var MonarchStackElementFactory = _MonarchStackElementFactory;
MonarchStackElementFactory._INSTANCE = new _MonarchStackElementFactory(CACHE_STACK_DEPTH);
var MonarchStackElement = class {
  constructor(parent, state) {
    this.parent = parent;
    this.state = state;
    this.depth = (this.parent ? this.parent.depth : 0) + 1;
  }
  static getStackElementId(element) {
    let result = "";
    while (element !== null) {
      if (result.length > 0) {
        result += "|";
      }
      result += element.state;
      element = element.parent;
    }
    return result;
  }
  static _equals(a, b) {
    while (a !== null && b !== null) {
      if (a === b) {
        return true;
      }
      if (a.state !== b.state) {
        return false;
      }
      a = a.parent;
      b = b.parent;
    }
    if (a === null && b === null) {
      return true;
    }
    return false;
  }
  get indent() {
    return this.state.lastIndexOf("	") - this.state.indexOf("	");
  }
  get scope() {
    return this.part(2);
  }
  get detail() {
    return this.part(2);
  }
  part(index) {
    return this.state.split(".")[index];
  }
  equals(other) {
    return MonarchStackElement._equals(this, other);
  }
  push(state) {
    return MonarchStackElementFactory.create(this, state);
  }
  pop() {
    return this.parent;
  }
  popall() {
    let result = this;
    while (result.parent) {
      result = result.parent;
    }
    return result;
  }
  switchTo(state) {
    return MonarchStackElementFactory.create(this.parent, state);
  }
};
var _MonarchLineStateFactory = class {
  static create(stack) {
    return this._INSTANCE.create(stack);
  }
  constructor(maxCacheDepth) {
    this._maxCacheDepth = maxCacheDepth;
    this._entries = Object.create(null);
  }
  create(stack) {
    if (stack !== null && stack.depth >= this._maxCacheDepth) {
      return new MonarchLineState(stack);
    }
    let stackElementId = MonarchStackElement.getStackElementId(stack);
    let result = this._entries[stackElementId];
    if (result) {
      return result;
    }
    result = new MonarchLineState(stack);
    this._entries[stackElementId] = result;
    return result;
  }
};
var MonarchLineStateFactory = _MonarchLineStateFactory;
MonarchLineStateFactory._INSTANCE = new _MonarchLineStateFactory(CACHE_STACK_DEPTH);
var MonarchLineState = class {
  constructor(stack) {
    this.stack = stack;
  }
  clone() {
    return MonarchLineStateFactory.create(this.stack);
  }
  equals(other) {
    if (!(other instanceof MonarchLineState)) {
      return false;
    }
    if (!this.stack.equals(other.stack)) {
      return false;
    }
    return true;
  }
};
var MonarchClassicTokensCollector = class {
  constructor() {
    this._tokens = [];
    this._language = null;
    this._lastToken = new Token(0, "start", "imba");
    this._lastTokenType = null;
  }
  enterMode(startOffset, modeId) {
    this._language = modeId;
  }
  emit(startOffset, type, stack) {
    if (this._lastTokenType === type && false) {
      console.log("add to last token", type);
      return this._lastToken;
    }
    let token = new Token(startOffset, type, this._language);
    this._lastTokenType = type;
    this._lastToken = token;
    this._tokens.push(token);
    return token;
  }
  finalize(endState) {
    return new TokenizationResult(this._tokens, endState);
  }
};
var MonarchTokenizer = class {
  constructor(modeId, lexer2) {
    this._modeId = modeId;
    this._lexer = lexer2;
    this._profile = false;
  }
  dispose() {
  }
  getLoadStatus() {
    return {loaded: true};
  }
  getInitialState() {
    let rootState = MonarchStackElementFactory.create(null, this._lexer.start);
    return MonarchLineStateFactory.create(rootState);
  }
  tokenize(line, lineState, offsetDelta) {
    let tokensCollector = new MonarchClassicTokensCollector();
    let endLineState = this._tokenize(line, lineState, offsetDelta, tokensCollector);
    return tokensCollector.finalize(endLineState);
  }
  _tokenize(line, lineState, offsetDelta, collector) {
    return this._myTokenize(line, lineState, offsetDelta, collector);
  }
  _safeRuleName(rule) {
    if (rule) {
      return rule.name;
    }
    return "(unknown)";
  }
  _rescope(from, to, tokens, toState) {
    let a = (from || "").split("-");
    let b = (to || "").split("-");
    if (from == to)
      return;
    let diff = 1;
    while (a[diff] && a[diff] == b[diff]) {
      diff++;
    }
    let level = a.length;
    while (level > diff) {
      tokens.push("pop." + a[--level] + "." + level);
    }
    while (b.length > diff) {
      let id = "push." + b[diff++] + "." + (diff - 1);
      if (toState) {
        let indent = statePart(toState, 1);
        id += "." + indent;
      }
      tokens.push(id);
    }
  }
  _myTokenize(line, lineState, offsetDelta, tokensCollector) {
    tokensCollector.enterMode(offsetDelta, this._modeId);
    const lineLength = line.length;
    let stack = lineState.stack;
    let lastToken = null;
    let pos = 0;
    let profile = this._profile;
    let groupMatching = null;
    let forceEvaluation = true;
    let append = [];
    let tries = 0;
    let rules = [];
    let rulesState = null;
    let hangPos = -1;
    while (forceEvaluation || pos < lineLength) {
      tries++;
      if (tries > 1e3) {
        if (pos == hangPos) {
          console.log("infinite recursion", pos, lineLength, stack, tokensCollector);
          throw "infinite recursion in tokenizer?";
        } else {
          hangPos = pos;
          tries = 0;
        }
      }
      const pos0 = pos;
      const stackLen0 = stack.depth;
      const groupLen0 = groupMatching ? groupMatching.groups.length : 0;
      const state = stack.state;
      let matches = null;
      let matched = null;
      let action = null;
      let rule = null;
      if (groupMatching) {
        matches = groupMatching.matches;
        const groupEntry = groupMatching.groups.shift();
        matched = groupEntry.matched;
        action = groupEntry.action;
        rule = groupMatching.rule;
        if (groupMatching.groups.length === 0) {
          groupMatching = null;
        }
      } else {
        if (!forceEvaluation && pos >= lineLength) {
          break;
        }
        forceEvaluation = false;
        rules = this._lexer.tokenizer[state];
        if (!rules) {
          rules = findRules(this._lexer, state);
          if (!rules) {
            throw createError(this._lexer, "tokenizer state is not defined: " + state);
          }
        }
        let restOfLine = line.substr(pos);
        for (const rule2 of rules) {
          if (rule2.string !== void 0) {
            if (restOfLine[0] === rule2.string) {
              matches = [rule2.string];
              matched = rule2.string;
              action = rule2.action;
              break;
            }
          } else if (pos === 0 || !rule2.matchOnlyAtLineStart) {
            if (profile) {
              rule2.stats.count++;
              let now = performance.now();
              matches = restOfLine.match(rule2.regex);
              rule2.stats.time += performance.now() - now;
              if (matches) {
                rule2.stats.hits++;
              }
            } else {
              matches = restOfLine.match(rule2.regex);
            }
            if (matches) {
              matched = matches[0];
              action = rule2.action;
              break;
            }
          }
        }
      }
      if (!matches) {
        matches = [""];
        matched = "";
      }
      if (!action) {
        if (pos < lineLength) {
          matches = [line.charAt(pos)];
          matched = matches[0];
        }
        action = this._lexer.defaultToken;
      }
      if (matched === null) {
        break;
      }
      pos += matched.length;
      while (isFuzzyAction(action) && isIAction(action) && action.test) {
        action = action.test(matched, matches, state, pos === lineLength);
      }
      let result = null;
      if (typeof action === "string" || Array.isArray(action)) {
        result = action;
      } else if (action.group) {
        result = action.group;
      } else if (action.token !== null && action.token !== void 0) {
        if (action.tokenSubst) {
          result = substituteMatches(this._lexer, action.token, matched, matches, state);
        } else {
          result = action.token;
        }
        if (action.goBack) {
          pos = Math.max(0, pos - action.goBack);
        }
        if (action.switchTo && typeof action.switchTo === "string") {
          let nextState = substituteMatches(this._lexer, action.switchTo, matched, matches, state);
          if (nextState[0] === "@") {
            nextState = nextState.substr(1);
          }
          if (!findRules(this._lexer, nextState)) {
            throw createError(this._lexer, "trying to switch to a state '" + nextState + "' that is undefined in rule: " + this._safeRuleName(rule));
          } else {
            let from = stack.scope;
            let to = statePart(nextState, 2);
            if (from !== to)
              this._rescope(from, to, append, nextState);
            stack = stack.switchTo(nextState);
          }
        } else if (action.transform && typeof action.transform === "function") {
          throw createError(this._lexer, "action.transform not supported");
        } else if (action.next) {
          if (action.next === "@push") {
            if (stack.depth >= this._lexer.maxStack) {
              throw createError(this._lexer, "maximum tokenizer stack size reached: [" + stack.state + "," + stack.parent.state + ",...]");
            } else {
              stack = stack.push(state);
            }
          } else if (action.next === "@pop") {
            if (stack.depth <= 1) {
              throw createError(this._lexer, "trying to pop an empty stack in rule: " + this._safeRuleName(rule));
            } else {
              let prev = stack;
              stack = stack.pop();
              let from = statePart(prev.state, 2);
              let to = statePart(stack.state, 2);
              if (from !== to)
                this._rescope(from, to, append, stack.state);
            }
          } else if (action.next === "@popall") {
            stack = stack.popall();
          } else {
            let nextState = substituteMatches(this._lexer, action.next, matched, matches, state);
            if (nextState[0] === "@") {
              nextState = nextState.substr(1);
            }
            let nextScope = statePart(nextState, 2);
            if (!findRules(this._lexer, nextState)) {
              throw createError(this._lexer, "trying to set a next state '" + nextState + "' that is undefined in rule: " + this._safeRuleName(rule));
            } else {
              if (nextScope != stack.scope)
                this._rescope(stack.scope || "", nextScope, append, nextState);
              stack = stack.push(nextState);
            }
          }
        }
        if (action.log && typeof action.log === "string") {
          log(this._lexer, this._lexer.languageId + ": " + substituteMatches(this._lexer, action.log, matched, matches, state));
        }
        if (action.mark) {
          tokensCollector.emit(pos0 + offsetDelta, action.mark, stack);
        }
      }
      if (result === null) {
        throw createError(this._lexer, "lexer rule has no well-defined action in rule: " + this._safeRuleName(rule));
      }
      if (Array.isArray(result)) {
        if (groupMatching && groupMatching.groups.length > 0) {
          throw createError(this._lexer, "groups cannot be nested: " + this._safeRuleName(rule));
        }
        if (matches.length !== result.length + 1) {
          throw createError(this._lexer, "matched number of groups does not match the number of actions in rule: " + this._safeRuleName(rule));
        }
        let totalLen = 0;
        for (let i = 1; i < matches.length; i++) {
          totalLen += matches[i].length;
        }
        if (totalLen !== matched.length) {
          throw createError(this._lexer, "with groups, all characters should be matched in consecutive groups in rule: " + this._safeRuleName(rule));
        }
        groupMatching = {
          rule,
          matches,
          groups: []
        };
        for (let i = 0; i < result.length; i++) {
          groupMatching.groups[i] = {
            action: result[i],
            matched: matches[i + 1]
          };
        }
        pos -= matched.length;
        continue;
      } else {
        if (result === "@rematch") {
          pos -= matched.length;
          matched = "";
          matches = null;
          result = "";
        }
        if (matched.length === 0) {
          if (lineLength === 0 || stackLen0 !== stack.depth || state !== stack.state || (!groupMatching ? 0 : groupMatching.groups.length) !== groupLen0) {
            if (typeof result == "string" && result)
              tokensCollector.emit(pos + offsetDelta, result, stack);
            while (append.length > 0) {
              tokensCollector.emit(pos + offsetDelta, append.shift(), stack);
            }
            continue;
          } else {
            throw createError(this._lexer, "no progress in tokenizer in rule: " + this._safeRuleName(rule));
          }
        }
        let tokenType = null;
        if (isString(result) && result.indexOf("@brackets") === 0) {
          let rest = result.substr("@brackets".length);
          let bracket = findBracket(this._lexer, matched);
          if (!bracket) {
            throw createError(this._lexer, "@brackets token returned but no bracket defined as: " + matched);
          }
          tokenType = sanitize(bracket.token + rest);
        } else {
          let token2 = result === "" ? "" : result + this._lexer.tokenPostfix;
          tokenType = sanitize(token2);
        }
        let token = tokensCollector.emit(pos0 + offsetDelta, tokenType, stack);
        token.stack = stack;
        if (lastToken && lastToken != token) {
          lastToken.value = line.slice(lastToken.offset - offsetDelta, pos0);
        }
        lastToken = token;
        while (append.length > 0) {
          tokensCollector.emit(pos + offsetDelta, append.shift(), stack);
        }
      }
    }
    if (lastToken && !lastToken.value) {
      lastToken.value = line.slice(lastToken.offset - offsetDelta);
    }
    return MonarchLineStateFactory.create(stack);
  }
};
function findBracket(lexer2, matched) {
  if (!matched) {
    return null;
  }
  matched = fixCase(lexer2, matched);
  let brackets = lexer2.brackets;
  for (const bracket of brackets) {
    if (bracket.open === matched) {
      return {token: bracket.token, bracketType: MonarchBracket.Open};
    } else if (bracket.close === matched) {
      return {token: bracket.token, bracketType: MonarchBracket.Close};
    }
  }
  return null;
}

// src/program/lexer.imba
function iter$__3(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
var φ12 = Symbol.for("#init");
var LexedLine = class {
  constructor($$ = null) {
    this[φ12]($$);
  }
  [φ12]($$ = null) {
    this.offset = $$ ? $$.offset : void 0;
    this.text = $$ ? $$.text : void 0;
    this.tokens = $$ ? $$.tokens : void 0;
    this.startState = $$ ? $$.startState : void 0;
    this.endState = $$ ? $$.endState : void 0;
  }
  clone(newOffset = this.offset) {
    let clones = [];
    let delta = newOffset - this.offset;
    for (let φ23 = 0, φ32 = iter$__3(this.tokens), φ42 = φ32.length; φ23 < φ42; φ23++) {
      let tok = φ32[φ23];
      let clone = tok.clone();
      clone.offset = tok.offset + delta;
      clones.push(clone);
    }
    ;
    return new LexedLine({
      offset: newOffset,
      startState: this.startState,
      endState: this.endState,
      text: this.text,
      tokens: clones
    });
  }
};
var compiled = compile("imba", grammar);
var lexer = new MonarchTokenizer("imba", compiled);

// src/program/types.imba
function iter$__4(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
var SemanticTokenTypes = [
  "comment",
  "string",
  "keyword",
  "number",
  "regexp",
  "operator",
  "namespace",
  "type",
  "struct",
  "class",
  "interface",
  "enum",
  "typeParameter",
  "function",
  "member",
  "macro",
  "variable",
  "parameter",
  "property",
  "label"
];
for (let index = 0, φ16 = iter$__4(SemanticTokenTypes), φ23 = φ16.length; index < φ23; index++) {
  let key = φ16[index];
  SemanticTokenTypes[key] = index;
}
var M = {
  Declaration: 1 << 0,
  Import: 1 << 1,
  Export: 1 << 2,
  Global: 1 << 3,
  ReadOnly: 1 << 4,
  Static: 1 << 5,
  Modification: 1 << 6,
  Deprecated: 1 << 7,
  Access: 1 << 8,
  Root: 1 << 9,
  Special: 1 << 10,
  Class: 1 << 11,
  Member: 1 << 12,
  Function: 1 << 13,
  Def: 1 << 14,
  Var: 1 << 15,
  Let: 1 << 16,
  Const: 1 << 17,
  Get: 1 << 18,
  Set: 1 << 19
};
var SemanticTokenModifiers = Object.keys(M).map(function(_0) {
  return _0.toLowerCase();
});
for (let φ32 = 0, φ42 = iter$__4(Object.keys(M)), φ52 = φ42.length; φ32 < φ52; φ32++) {
  let k = φ42[φ32];
  M[k.toLowerCase()] = M[k];
}
var CompletionTypes = {
  Keyword: 1 << 0,
  Access: 1 << 1,
  Key: 1 << 2,
  TagName: 1 << 3,
  TagEvent: 1 << 4,
  TagFlag: 1 << 5,
  TagProp: 1 << 6,
  TagEventModifier: 1 << 7,
  Value: 1 << 8,
  Path: 1 << 9,
  StyleProp: 1 << 10,
  StyleValue: 1 << 11,
  Type: 1 << 12,
  StyleModifier: 1 << 13,
  StyleSelector: 1 << 14,
  VarName: 1 << 15,
  ImportName: 1 << 16
};
var KeywordTypes = {
  Keyword: 1 << 0,
  Root: 1 << 1,
  Class: 1 << 2,
  Block: 1 << 3
};
var Keywords = {
  and: KeywordTypes.Block,
  await: KeywordTypes.Block,
  begin: KeywordTypes.Block,
  break: KeywordTypes.Block,
  by: KeywordTypes.Block,
  case: KeywordTypes.Block,
  catch: KeywordTypes.Block,
  class: KeywordTypes.Block,
  const: KeywordTypes.Block,
  continue: KeywordTypes.Block,
  css: KeywordTypes.Class | KeywordTypes.Root,
  debugger: KeywordTypes.Block,
  def: KeywordTypes.Class | KeywordTypes.Block,
  get: KeywordTypes.Class,
  set: KeywordTypes.Class,
  delete: KeywordTypes.Block,
  do: KeywordTypes.Block,
  elif: KeywordTypes.Block,
  else: KeywordTypes.Block,
  export: KeywordTypes.Root,
  extends: KeywordTypes.Block,
  false: KeywordTypes.Block,
  finally: KeywordTypes.Block,
  for: KeywordTypes.Block,
  if: KeywordTypes.Block,
  import: KeywordTypes.Root,
  in: KeywordTypes.Block,
  instanceof: KeywordTypes.Block,
  is: KeywordTypes.Block,
  isa: KeywordTypes.Block,
  isnt: KeywordTypes.Block,
  let: KeywordTypes.Block,
  loop: KeywordTypes.Block,
  module: KeywordTypes.Block,
  nil: KeywordTypes.Block,
  no: KeywordTypes.Block,
  not: KeywordTypes.Block,
  null: KeywordTypes.Block,
  of: KeywordTypes.Block,
  or: KeywordTypes.Block,
  require: KeywordTypes.Block,
  return: KeywordTypes.Block,
  self: KeywordTypes.Block,
  static: KeywordTypes.Block | KeywordTypes.Class,
  super: KeywordTypes.Block,
  switch: KeywordTypes.Block,
  tag: KeywordTypes.Root,
  then: KeywordTypes.Block,
  this: KeywordTypes.Block,
  throw: KeywordTypes.Block,
  true: KeywordTypes.Block,
  try: KeywordTypes.Block,
  typeof: KeywordTypes.Block,
  undefined: KeywordTypes.Block,
  unless: KeywordTypes.Block,
  until: KeywordTypes.Block,
  var: KeywordTypes.Block,
  when: KeywordTypes.Block,
  while: KeywordTypes.Block,
  yes: KeywordTypes.Block
};
var SymbolKind = {
  File: 1,
  Module: 2,
  Namespace: 3,
  Package: 4,
  Class: 5,
  Method: 6,
  Property: 7,
  Field: 8,
  Constructor: 9,
  Enum: 10,
  Interface: 11,
  Function: 12,
  Variable: 13,
  Constant: 14,
  String: 15,
  Number: 16,
  Boolean: 17,
  Array: 18,
  Object: 19,
  Key: 20,
  Null: 21,
  EnumMember: 22,
  Struct: 23,
  Event: 24,
  Operator: 25,
  TypeParameter: 26
};
for (let φ6 = 0, φ72 = Object.keys(SymbolKind), φ82 = φ72.length, k, v; φ6 < φ82; φ6++) {
  k = φ72[φ6];
  v = SymbolKind[k];
  SymbolKind[v] = k;
}

// src/program/symbol.imba
function iter$__5(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
var φ13 = Symbol.for("#init");
var φ5 = Symbol.for("#datatype");
var SymbolFlags = {
  None: 0,
  ConstVariable: 1 << 0,
  LetVariable: 1 << 1,
  Property: 1 << 2,
  EnumMember: 1 << 3,
  Function: 1 << 4,
  Class: 1 << 5,
  LocalComponent: 1 << 6,
  GlobalComponent: 1 << 7,
  RegularEnum: 1 << 8,
  ValueModule: 1 << 9,
  Parameter: 1 << 10,
  TypeLiteral: 1 << 11,
  ObjectLiteral: 1 << 12,
  Method: 1 << 13,
  Constructor: 1 << 14,
  GetAccessor: 1 << 15,
  SetAccessor: 1 << 16,
  Signature: 1 << 17,
  TypeParameter: 1 << 18,
  TypeAlias: 1 << 19,
  ExportValue: 1 << 20,
  Alias: 1 << 21,
  Prototype: 1 << 22,
  ExportStar: 1 << 23,
  Optional: 1 << 24,
  IsSpecial: 1 << 27,
  IsImport: 1 << 28,
  IsStatic: 1 << 29,
  IsGlobal: 1 << 30,
  IsRoot: 1 << 31
};
SymbolFlags.Component = SymbolFlags.LocalComponent | SymbolFlags.GlobalComponent;
SymbolFlags.Variable = SymbolFlags.LetVariable | SymbolFlags.ConstVariable | SymbolFlags.Parameter;
SymbolFlags.Accessor = SymbolFlags.GetAccessor | SymbolFlags.SetAccessor;
SymbolFlags.ClassMember = SymbolFlags.Method | SymbolFlags.Accessor | SymbolFlags.Property;
SymbolFlags.Scoped = SymbolFlags.Function | SymbolFlags.Variable | SymbolFlags.Class | SymbolFlags.Enum | SymbolFlags.LocalComponent;
SymbolFlags.Type = SymbolFlags.Component | SymbolFlags.Class;
SymbolFlags.GlobalVar = SymbolFlags.ConstVariable | SymbolFlags.IsGlobal;
SymbolFlags.SpecialVar = SymbolFlags.ConstVariable | SymbolFlags.IsSpecial;
var Conversions = [
  ["entity.name.component.local", 0, SymbolFlags.LocalComponent],
  ["entity.name.component.global", 0, SymbolFlags.GlobalComponent],
  ["entity.name.function", 0, SymbolFlags.Function],
  ["entity.name.class", 0, SymbolFlags.Class],
  ["entity.name.constructor", 0, SymbolFlags.Method],
  ["entity.name.def", 0, SymbolFlags.Method],
  ["entity.name.get", 0, SymbolFlags.GetAccessor],
  ["entity.name.set", 0, SymbolFlags.SetAccessor],
  ["field", 0, SymbolFlags.Property],
  ["decl-let", 0, SymbolFlags.LetVariable],
  ["decl-for-index", 0, SymbolFlags.LetVariable, {datatype: "\\number"}],
  ["decl-for", 0, SymbolFlags.LetVariable, {kind: "for"}],
  ["decl-var", 0, SymbolFlags.LetVariable],
  ["decl-param", 0, SymbolFlags.Parameter],
  ["decl-const", 0, SymbolFlags.ConstVariable],
  ["decl-import", 0, SymbolFlags.ConstVariable | SymbolFlags.IsImport]
];
var ConversionCache = {};
var Sym = class {
  [φ13]($$ = null) {
    var φ23;
    this.value = $$ ? $$.value : void 0;
    this.body = $$ && (φ23 = $$.body) !== void 0 ? φ23 : null;
  }
  static typeMatch(type) {
    if (ConversionCache[type] != void 0) {
      return ConversionCache[type];
    }
    ;
    for (let i = 0, φ32 = iter$__5(Conversions), φ42 = φ32.length; i < φ42; i++) {
      let [strtest, modtest, flags, o] = φ32[i];
      if (type.indexOf(strtest) >= 0) {
        return ConversionCache[type] = Conversions[i];
      }
      ;
    }
    ;
    return null;
  }
  static forToken(tok, type, mods = 0) {
    let match = this.typeMatch(type);
    if (match) {
      let sym = new this(match[2], tok.value, tok, match[3]);
      return sym;
    }
    ;
    return null;
  }
  constructor(flags, name, node, desc = null) {
    this[φ13]();
    this.flags = flags;
    this.name = name;
    this.node = node;
    this.desc = desc;
  }
  get importSource() {
    if (!this.importedΦ) {
      return null;
    }
    ;
    let ctx = this.node.context.closest("imports");
    return ctx.sourcePath;
  }
  get exportName() {
    if (this.node.prev.match("keyword.as")) {
      return this.node.prev.prev.value;
    } else if (this.node.match(".default")) {
      return "default";
    } else {
      return this.node.value;
    }
    ;
  }
  get importInfo() {
    if (!this.importedΦ) {
      return null;
    }
    ;
    let ctx = this.node.context.closest("imports");
    return {
      exportName: this.exportName,
      name: this.node.value,
      isTypeOnly: ctx.isTypeOnly,
      path: ctx.sourcePath
    };
  }
  get datatype() {
    var _a;
    let m;
    let type = this.desc && this.desc.datatype;
    if (type) {
      return type;
    }
    ;
    if (this[φ5]) {
      return this[φ5];
    }
    ;
    let next = this.node && this.node.nextNode;
    if (next && next.type == "type") {
      return next;
    }
    ;
    let scope = this.scope;
    if (scope && ((_a = this.desc) == null ? void 0 : _a.kind) == "for") {
      let typ = scope.doc.getDestructuredPath(this.node, [[scope.expression, "__@iterable"]]);
      return this[φ5] || (this[φ5] = typ);
    }
    ;
    if (m = this.importInfo) {
      return m;
    }
    ;
    return null;
  }
  get staticΦ() {
    return this.node && this.node.mods & M.Static;
  }
  get itervarΦ() {
    return this.node && this.node.match(".decl-for");
  }
  get variableΦ() {
    return this.flags & SymbolFlags.Variable;
  }
  get parameterΦ() {
    return this.flags & SymbolFlags.Parameter;
  }
  get memberΦ() {
    return this.flags & SymbolFlags.ClassMember;
  }
  get scopedΦ() {
    return this.flags & SymbolFlags.Scoped;
  }
  get typeΦ() {
    return this.flags & SymbolFlags.Type;
  }
  get globalΦ() {
    return this.flags & SymbolFlags.IsGlobal;
  }
  get importedΦ() {
    return this.flags & SymbolFlags.IsImport;
  }
  get componentΦ() {
    return this.flags & SymbolFlags.Component;
  }
  get escapedName() {
    return this.name;
  }
  get scope() {
    var _a, _b;
    return (_b = (_a = this.node) == null ? void 0 : _a.context) == null ? void 0 : _b.scope;
  }
  addReference(node) {
    this.references || (this.references = []);
    this.references.push(node);
    node.symbol = this;
    return this;
  }
  dereference(tok) {
    let idx = this.references.indexOf(tok);
    if (idx >= 0) {
      tok.symbol = null;
      this.references.splice(idx, 1);
    }
    ;
    return this;
  }
  get kind() {
    if (this.variableΦ) {
      return SymbolKind.Variable;
    } else if (this.flags & SymbolFlags.Class) {
      return SymbolKind.Class;
    } else if (this.flags & SymbolFlags.Component) {
      return SymbolKind.Class;
    } else if (this.flags & SymbolFlags.Property) {
      return SymbolKind.Field;
    } else if (this.flags & SymbolFlags.Method) {
      if (this.escapedName == "constructor") {
        return SymbolKind.Constructor;
      } else {
        return SymbolKind.Method;
      }
      ;
    } else if (this.flags & SymbolFlags.Function) {
      return SymbolKind.Function;
    } else {
      return SymbolKind.Method;
    }
    ;
  }
  get semanticKind() {
    if (this.flags & SymbolFlags.Parameter) {
      return "parameter";
    } else if (this.variableΦ) {
      return "variable";
    } else if (this.typeΦ) {
      return "type";
    } else if (this.flags & SymbolFlags.Function) {
      return "function";
    } else if (this.memberΦ) {
      return "member";
    } else if (this.componentΦ) {
      return "component";
    } else {
      return "variable";
    }
    ;
  }
  get semanticFlags() {
    let mods = 0;
    if (this.flags & SymbolFlags.ConstVariable) {
      mods |= M.ReadOnly;
    }
    ;
    if (this.staticΦ) {
      mods |= M.Static;
    }
    ;
    if (this.flags & SymbolFlags.IsImport) {
      mods |= M.Import;
    }
    ;
    if (this.flags & SymbolFlags.IsGlobal) {
      mods |= M.Global;
    }
    ;
    if (this.flags & SymbolFlags.IsRoot) {
      mods |= M.Root;
    }
    ;
    if (this.flags & SymbolFlags.IsSpecial) {
      mods |= M.Special;
    }
    ;
    return mods;
  }
};

// src/program/scope.imba
function iter$__6(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
var φ14 = Symbol.for("#init");
var Globals = {
  global: {datatype: "globalThis"},
  imba: {datatype: "globalThis.imba"},
  module: {},
  window: {datatype: "globalThis.window"},
  document: {datatype: "globalThis.document"},
  exports: {},
  console: {datatype: "globalThis.console"},
  process: {datatype: "globalThis.process"},
  parseInt: {datatype: "globalThis.parseInt"},
  parseFloat: {datatype: "globalThis.parseFloat"},
  setTimeout: {datatype: "globalThis.setTimeout"},
  setInterval: {datatype: "globalThis.setInterval"},
  setImmediate: {datatype: "globalThis.setImmediate"},
  clearTimeout: {datatype: "globalThis.clearTimeout"},
  clearInterval: {datatype: "globalThis.clearInterval"},
  clearImmediate: {datatype: "globalThis.clearImmediate"},
  globalThis: {datatype: "globalThis"},
  isNaN: {datatype: "globalThis.isNaN"},
  isFinite: {datatype: "globalThis.isFinite"},
  __dirname: {datatype: "\\string"},
  __filename: {datatype: "\\string"},
  __realname: {datatype: "\\string"}
};
var Node = class {
  [φ14]($$ = null) {
    var φ23;
    this.type = $$ && (φ23 = $$.type) !== void 0 ? φ23 : "";
    this.start = $$ ? $$.start : void 0;
    this.end = $$ ? $$.end : void 0;
    this.parent = $$ ? $$.parent : void 0;
  }
  static build(doc, tok, scope, typ, types) {
    return new this(doc, tok, scope, typ, types);
  }
  constructor(doc, token, parent, type) {
    this[φ14]();
    this.doc = doc;
    this.start = token;
    this.end = null;
    this.type = type;
    this.parent = parent;
    this.$name = null;
    token.scope = this;
  }
  pop(end) {
    this.end = end;
    end.start = this.start;
    end.pops = this;
    this.start.end = end;
    this.visit();
    return this.parent;
  }
  find(pattern) {
    return this.findChildren(pattern, true)[0];
  }
  get childNodes() {
    let nodes = this.doc.getNodesInScope(this);
    return nodes;
  }
  findChildren(pattern, returnFirst = false) {
    let found = [];
    let tok = this.start;
    while (tok) {
      if (returnFirst && found.length) {
        return found;
      }
      ;
      if (tok.scope && tok.scope != this) {
        if (tok.scope.match(pattern)) {
          found.push(tok.scope);
        }
        ;
        tok = tok.scope.next;
        continue;
      }
      ;
      if (tok.match(pattern)) {
        found.push(tok);
      }
      ;
      if (tok == this.end) {
        break;
      }
      ;
      tok = tok.next;
    }
    ;
    return found;
  }
  closest(ref) {
    if (this.match(ref)) {
      return this;
    }
    ;
    return this.parent ? this.parent.closest(ref) : null;
  }
  visit() {
    return this;
  }
  get memberΦ() {
    return false;
  }
  get topΦ() {
    return false;
  }
  get selfScope() {
    return this.memberΦ || this.topΦ ? this : this.parent.selfScope;
  }
  get name() {
    return this.$name || "";
  }
  get value() {
    return this.doc.content.slice(this.start.offset, this.next ? this.next.offset : -1);
  }
  get next() {
    return this.end ? this.end.next : null;
  }
  get prev() {
    return this.start ? this.start.prev : null;
  }
  match(query) {
    if (typeof query == "string") {
      return this.type == query;
    } else if (query instanceof RegExp) {
      return query.test(this.type);
    } else if (query instanceof Function) {
      return query(this);
    }
    ;
    return true;
  }
};
var Group = class extends Node {
  constructor(doc, token, parent, type, parts = []) {
    super(doc, token, parent, type);
  }
  get scope() {
    return this.parent.scope;
  }
  get varmap() {
    return this.parent.varmap;
  }
  register(symbol) {
    return this.parent.register(symbol);
  }
  lookup(...params) {
    return this.parent.lookup(...params);
  }
};
var ValueNode = class extends Group {
};
var StringNode = class extends Group {
};
var StyleNode = class extends Group {
  get properties() {
    return this.findChildren("styleprop");
  }
};
var StyleRuleNode = class extends Group {
};
var Scope = class extends Node {
  constructor(doc, token, parent, type, parts = []) {
    super(doc, token, parent, type);
    this.children = [];
    this.entities = [];
    this.refs = [];
    this.varmap = Object.create(parent ? parent.varmap : {});
    if (this instanceof Root) {
      for (let φ32 = 0, φ42 = Object.keys(Globals), φ52 = φ42.length, key, val; φ32 < φ52; φ32++) {
        key = φ42[φ32];
        val = Globals[key];
        let tok = {value: key, offset: -1, mods: 0};
        this.varmap[key] = new Sym(SymbolFlags.GlobalVar, key, tok, val);
      }
      ;
    }
    ;
    this.indent = parts[3] && parts[3][0] == "	" ? parts[3].length : 0;
    this.setup();
    return this;
  }
  setup() {
    if (this.handlerΦ) {
      this.varmap.e = new Sym(SymbolFlags.SpecialVar, "e", null, "eventReference");
    }
    ;
    if (this.classΦ || this.propertyΦ) {
      this.ident = this.token = prevToken(this.start, "entity.");
      if (this.ident) {
        this.ident.body = this;
      }
      ;
      if (this.ident && this.ident.type == "entity.name.def.render") {
        this.$name = "render";
        if (this.ident.symbol) {
          return this.ident.symbol.name = "render";
        }
        ;
      }
      ;
    }
    ;
  }
  get selfPath() {
    let path = this.path;
    if (this.propertyΦ) {
      return path.slice(0, path.lastIndexOf("."));
    }
    ;
    return path;
  }
  get path() {
    let par = this.parent ? this.parent.path : "";
    if (this.propertyΦ) {
      let sep = this.staticΦ ? "." : ".prototype.";
      return this.parent ? "" + this.parent.path + sep + this.name : this.name;
    }
    ;
    if (this.componentΦ) {
      if (this.name[0] == this.name[0].toLowerCase()) {
        return this.name.replace(/\-/g, "_") + "$$TAG$$";
      } else {
        return this.name;
      }
      ;
    }
    ;
    if (this.classΦ) {
      return this.name;
    }
    ;
    return par;
  }
  get allowedKeywordTypes() {
    if (this.classΦ) {
      return KeywordTypes.Class;
    } else if (this.rootΦ) {
      return KeywordTypes.Root | KeywordTypes.Block;
    } else {
      return KeywordTypes.Block;
    }
    ;
  }
  get componentΦ() {
    return !!this.type.match(/^component/);
  }
  get rootΦ() {
    return this instanceof Root;
  }
  get topΦ() {
    return this instanceof Root;
  }
  get classΦ() {
    return !!this.type.match(/^class/) || this.componentΦ;
  }
  get defΦ() {
    return !!this.type.match(/def|get|set/);
  }
  get staticΦ() {
    return !!(this.ident && this.ident.mods & M.Static);
  }
  get handlerΦ() {
    return !!this.type.match(/handler|spy/);
  }
  get memberΦ() {
    return !!this.type.match(/def|get|set/);
  }
  get propertyΦ() {
    return !!this.type.match(/def|get|set|field/);
  }
  get flowΦ() {
    return !!this.type.match(/if|else|elif|unless|for|while|until/);
  }
  get closureΦ() {
    return !!this.type.match(/class|component|def|get|set|do/);
  }
  get scope() {
    return this;
  }
  get name() {
    return this.$name || (this.ident ? this.ident.value : "");
  }
  visit() {
    return this;
  }
  register(symbol) {
    if (symbol.scopedΦ) {
      this.varmap[symbol.name] = symbol;
      if (this.rootΦ) {
        symbol.flags |= SymbolFlags.IsRoot;
      }
      ;
    }
    ;
    return symbol;
  }
  lookup(token, kind = SymbolFlags.Scoped) {
    let variable;
    let name = token.value;
    if (name[name.length - 1] == "!") {
      name = name.slice(0, -1);
    }
    ;
    if (variable = this.varmap[name]) {
      return variable;
    }
    ;
    return null;
  }
  toOutline() {
    return {
      kind: this.type,
      name: this.name,
      children: [],
      span: this.ident ? this.ident.span : this.start.span
    };
  }
};
var Root = class extends Scope {
};
var ForScope = class extends Scope {
  get expression() {
    let kw = this.find("keyword.in keyword.of");
    return kw.next.next;
  }
  get forvars() {
    return Object.values(this.varmap).filter(function(_0) {
      return _0.itervarΦ;
    });
  }
};
var WeakScope = class extends Scope {
  register(symbol) {
    return this.parent.register(symbol);
  }
  lookup(...params) {
    return this.parent.lookup(...params);
  }
};
var FieldScope = class extends Scope {
  get selfScope() {
    return this;
  }
};
var SelectorNode = class extends Group {
};
var StylePropKey = class extends Group {
  get propertyName() {
    var _a;
    if (this.start.next.match("style.property.name")) {
      return this.start.next.value;
    } else {
      return (_a = this.parent.prevProperty) == null ? void 0 : _a.propertyName;
    }
    ;
  }
  get modifier() {
    if (this.start.next.match("style.property.modifier")) {
      return this.start.next.value;
    }
    ;
  }
  get styleValue() {
    return true;
  }
};
var StylePropValue = class extends Group {
  get key() {
    return this.parent.key;
  }
  get propertyName() {
    return this.parent.propertyName;
  }
  get modifier() {
    return this.parent.modifier;
  }
};
var StylePropNode = class extends Group {
  get key() {
    return this.find("stylepropkey");
  }
  get prevProperty() {
    if (this.start.prev.pops) {
      return this.start.prev.pops;
    }
    ;
    return null;
  }
  get propertyName() {
    var _a;
    return (_a = this.key) == null ? void 0 : _a.propertyName;
  }
  get modifier() {
    var _a;
    return (_a = this.key) == null ? void 0 : _a.modifier;
  }
};
var StyleInterpolation = class extends Group {
};
var PathNode = class extends Group {
  get innerText() {
    return this.value.slice(1, -1);
  }
};
var TagNode = class extends Group {
  get name() {
    let name = this.findChildren("tag.name").join("");
    return name == "self" ? this.closest("component").name : name;
  }
  get localΦ() {
    return this.name[0] == this.name[0].toUpperCase();
  }
  get tagName() {
    return this.name;
  }
  get parentTag() {
    var _a;
    return (_a = this.closest("tagcontent")) == null ? void 0 : _a.ownerTag;
  }
  get ancestorTags() {
    var _a;
    return (_a = this.closest("tagcontent")) == null ? void 0 : _a.ownerTags;
  }
  get ancestorPath() {
    return this.ancestorTags.map(function(_0) {
      return _0.tagName;
    }).join(".");
  }
  get pathName() {
    return "<" + this.name + ">";
  }
  get outline() {
    return this.findChildren(/tag\.(reference|name|id|white|flag|event(?!\-))/).join("");
  }
};
var TagAttrNode = class extends Group {
  get propertyName() {
    if (this.start.next.match("tag.attr")) {
      return this.start.next.value;
    } else {
      return "";
    }
    ;
  }
  get tagName() {
    return this.parent.name;
  }
};
var TagAttrValueNode = class extends Group {
  get propertyName() {
    return this.parent.propertyName;
  }
  get tagName() {
    return this.parent.tagName;
  }
};
var TagContent = class extends WeakScope {
  get ownerTag() {
    return this.start.prev.pops;
  }
  get ownerTags() {
    let els = [this.ownerTag];
    let el;
    while (el = els[0].parentTag) {
      els.unshift(el);
    }
    ;
    return els;
  }
};
var Listener = class extends Group {
  get name() {
    return this.findChildren("tag.event.name").join("").replace("@", "");
  }
};
var ParensNode = class extends Group {
};
var BracketsNode = class extends Group {
  static build(doc, tok, scope, typ, types) {
    let cls = this;
    let chr = doc.content[tok.offset - 1];
    if (!chr || " [{(|=&-;\n	:/*%+-".indexOf(chr) >= 0) {
      typ = "array";
      cls = ArrayNode;
    } else {
      typ = "index";
      cls = IndexNode;
    }
    ;
    return new cls(doc, tok, scope, typ, types);
  }
};
var BracesNode = class extends Group {
};
var SpecifiersNode = class extends BracesNode {
};
var ArrayNode = class extends BracketsNode {
  get delimiters() {
    return this.childNodes.filter(function(_0) {
      return _0.match("delimiter");
    });
  }
  indexOfNode(node) {
    var φ6;
    let delims = this.delimiters;
    let index = 0;
    φ6 = 0;
    for (let delim of iter$__6(delims)) {
      let i = φ6++;
      if (node.offset > delim.offset) {
        index++;
      }
      ;
    }
    ;
    return index;
  }
};
var IndexNode = class extends BracketsNode {
};
var TypeAnnotationNode = class extends Group {
  constructor() {
    super(...arguments);
    this.prev.datatype = this;
  }
  toString() {
    return this.value;
  }
};
var InterpolatedValueNode = class extends Group {
};
var ObjectNode = class extends BracesNode {
};
var ImportsNode = class extends Group {
  get isTypeOnly() {
    return this.start.prev.match("keyword.type");
  }
  get sourcePath() {
    let path = this.childNodes.find(function(_0) {
      return _0.match("path");
    });
    return path == null ? void 0 : path.innerText;
  }
  get specifiers() {
    return this.childNodes.find(function(_0) {
      return _0.match("specifiers");
    });
  }
  get default() {
    return this.childNodes.find(function(_0) {
      return _0.match(".default");
    });
  }
  get namespace() {
    return this.childNodes.find(function(_0) {
      return _0.match(".ns");
    });
  }
};
var ScopeTypeMap = {
  style: StyleNode,
  array: BracketsNode,
  stylerule: StyleRuleNode,
  sel: SelectorNode,
  path: PathNode,
  value: ValueNode,
  tag: TagNode,
  forscope: ForScope,
  field: FieldScope,
  type: TypeAnnotationNode,
  parens: ParensNode,
  brackets: BracketsNode,
  object: ObjectNode,
  braces: BracesNode,
  specifiers: SpecifiersNode,
  string: StringNode,
  tagattr: TagAttrNode,
  imports: ImportsNode,
  interpolation: InterpolatedValueNode,
  tagattrvalue: TagAttrValueNode,
  tagcontent: TagContent,
  listener: Listener,
  styleinterpolation: StyleInterpolation,
  styleprop: StylePropNode,
  stylepropkey: StylePropKey,
  stylevalue: StylePropValue,
  args: ParensNode
};

// src/program/document.imba
function extend$__(target, ext) {
  var descriptors = Object.getOwnPropertyDescriptors(ext);
  Object.defineProperties(target, descriptors);
  return target;
}
function iter$__7(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
var φ15 = Symbol.for("#ins");
var φ22 = Symbol.for("#del");
var φ7 = Symbol.for("#version");
var φ8 = Symbol.for("#multiline");
var φ9 = Symbol.for("#significant");
var φ10 = Symbol.for("#body");
var φ20 = Symbol.for("#lexed");
var Extend$Token$af = class {
  get node() {
    if (this.scope && this.scope.start == this) {
      return this.scope;
    }
    ;
    if (this.pops) {
      return this.pops;
    }
    ;
    return this;
  }
  get nextNode() {
    var _a;
    return (_a = this.next) == null ? void 0 : _a.node;
  }
  get prevNode() {
    var _a;
    return (_a = this.prev) == null ? void 0 : _a.node;
  }
};
extend$__(Token.prototype, Extend$Token$af.prototype);
var ImbaDocument = class {
  static tmp(content) {
    return new this("file://temporary.imba", "imba", 0, content);
  }
  static from(uri, languageId, version, content) {
    return new this(uri, languageId, version, content);
  }
  constructor(uri, languageId, version, content) {
    this.uri = uri;
    this.languageId = languageId;
    this.version = version;
    this.content = content;
    this.connection = null;
    this.lineTokens = [];
    this.isLegacy = languageId == "imba1" || uri && uri.match(/\.imba1$/);
    this.head = this.seed = new Token(0, "eol", "imba");
    this.initialState = lexer.getInitialState();
    this.seed.stack = lexer.getInitialState();
    this.history = [];
    this.lexer = lexer;
    this.versionToHistoryMap = {};
    this.versionToHistoryMap[version] = -1;
    if (content && content.match(/^\#[^\n]+imba1/m)) {
      this.isLegacy = true;
    }
    ;
  }
  log(...params) {
    return console.log(...params);
  }
  get lineCount() {
    return this.lineOffsets.length;
  }
  get lineOffsets() {
    return this._lineOffsets || (this._lineOffsets = computeLineOffsets(this.content, true));
  }
  getText(range = null) {
    if (range) {
      let start = this.offsetAt(range.start);
      let end = this.offsetAt(range.end);
      return this.content.substring(start, end);
    }
    ;
    return this.content;
  }
  getLineText(line) {
    let start = this.lineOffsets[line];
    let end = this.lineOffsets[line + 1];
    return this.content.substring(start, end).replace(/[\r\n]/g, "");
  }
  positionAt(offset) {
    if (offset instanceof Position) {
      return offset;
    }
    ;
    if (typeof offset == "object") {
      offset = offset.offset;
    }
    ;
    offset = Math.max(Math.min(offset, this.content.length), 0);
    let lineOffsets = this.lineOffsets;
    let low = 0;
    let high = lineOffsets.length;
    if (high === 0) {
      return new Position(0, offset, offset, this.version);
    }
    ;
    while (low < high) {
      let mid = Math.floor((low + high) / 2);
      if (lineOffsets[mid] > offset) {
        high = mid;
      } else {
        low = mid + 1;
      }
      ;
    }
    ;
    let line = low - 1;
    return new Position(line, offset - lineOffsets[line], offset, this.version);
  }
  offsetAt(position) {
    if (position.offset) {
      return position.offset;
    }
    ;
    let lineOffsets = this.lineOffsets;
    if (position.line >= lineOffsets.length) {
      return this.content.length;
    } else if (position.line < 0) {
      return 0;
    }
    ;
    let lineOffset = lineOffsets[position.line];
    let nextLineOffset = position.line + 1 < lineOffsets.length ? lineOffsets[position.line + 1] : this.content.length;
    return position.offset = Math.max(Math.min(lineOffset + position.character, nextLineOffset), lineOffset);
  }
  rangeAt(start, end = start) {
    return new Range(this.positionAt(start), this.positionAt(end));
  }
  overwrite(body, newVersion) {
    this.version = newVersion || this.version + 1;
    this.content = body;
    this._lineOffsets = null;
    this.invalidateFromLine(0);
    return this;
  }
  update(changes, version) {
    if (version == void 0) {
      version = this.version + 1;
    }
    ;
    let edits = [];
    let isSignificant = false;
    edits[φ15] = "";
    edits[φ22] = "";
    for (let i = 0, φ32 = iter$__7(changes), φ6 = φ32.length; i < φ6; i++) {
      let change = φ32[i];
      if (editIsFull(change)) {
        this.overwrite(change.text, version);
        edits.push([0, this.content.length, change.text]);
        isSignificant = true;
        continue;
      }
      ;
      let range = getWellformedRange(change.range);
      let startOffset = this.offsetAt(range.start);
      let endOffset = this.offsetAt(range.end);
      change.range = range;
      change.offset = startOffset;
      change.length = endOffset - startOffset;
      range.start.offset = startOffset;
      range.end.offset = endOffset;
      let remLength = endOffset - startOffset;
      let prevText = remLength ? this.content.slice(startOffset, endOffset) : "";
      edits[φ22] += prevText;
      edits[φ15] += change.text || "";
      this.applyEdit(change, version, changes);
      edits.push([startOffset, endOffset - startOffset, change.text || "", prevText]);
      let startLine = Math.max(range.start.line, 0);
      let endLine = Math.max(range.end.line, 0);
      let lineOffsets = this.lineOffsets;
      let addedLineOffsets = computeLineOffsets(change.text, false, startOffset);
      if (endLine - startLine === addedLineOffsets.length) {
        for (let k = 0, φ42 = iter$__7(addedLineOffsets), φ52 = φ42.length; k < φ52; k++) {
          let added = φ42[k];
          lineOffsets[k + startLine + 1] = addedLineOffsets[i];
        }
        ;
      } else {
        if (addedLineOffsets.length < 1e4) {
          lineOffsets.splice.apply(lineOffsets, [startLine + 1, endLine - startLine].concat(addedLineOffsets));
        } else {
          this._lineOffsets = lineOffsets = lineOffsets.slice(0, startLine + 1).concat(addedLineOffsets, lineOffsets.slice(endLine + 1));
        }
        ;
      }
      ;
      let diff = change.text.length - (endOffset - startOffset);
      if (diff !== 0) {
        let k = startLine + 1 + addedLineOffsets.length;
        while (k < lineOffsets.length) {
          lineOffsets[k] = lineOffsets[k] + diff;
          k++;
        }
        ;
      }
      ;
    }
    ;
    this.history.push(edits);
    edits[φ7] = version;
    let prevEdits = this.history[this.history.length - 2];
    let changeStr = edits[φ15] + edits[φ22];
    if (changeStr.indexOf("\n") >= 0) {
      edits[φ8] = true;
      if (prevEdits && !prevEdits[φ8]) {
        edits[φ9] = true;
        edits[φ10] = this.content;
      }
      ;
    }
    ;
    this.versionToHistoryMap[version] = this.history.length - 1;
    return this.updated(changes, version);
  }
  get lastSignificantVersion() {
    let i = this.history.length;
    while (i > 0) {
      let edits = this.history[--i];
      if (edits && edits[φ9]) {
        return edits[φ7];
      }
      ;
    }
    ;
    return null;
  }
  editsSinceVersion(version) {
    let from = this.versionToHistoryMap[version];
    let edits = [];
    for (let φ11 = 0, φ122 = iter$__7(this.history.slice(from + 1)), φ132 = φ122.length; φ11 < φ132; φ11++) {
      let item = φ122[φ11];
      edits.push(...item);
    }
    ;
    return edits;
  }
  contentAtVersion(version) {
    let nr = this.versionToHistoryMap[version];
    let edits = this.history[nr];
    if (edits[φ10] != void 0) {
      return edits[φ10];
    }
    ;
    return null;
  }
  offsetAtVersion(fromOffset, fromVersion, toVersion = this.version, stickyStart = false) {
    let from = this.versionToHistoryMap[fromVersion];
    let to = this.versionToHistoryMap[toVersion];
    let offset = fromOffset;
    let modified = false;
    if (from < to) {
      while (from < to) {
        let edits = this.history[++from];
        for (let φ142 = 0, φ152 = iter$__7(edits), φ16 = φ152.length; φ142 < φ16; φ142++) {
          let [start, len, text] = φ152[φ142];
          if (start > offset) {
            continue;
          }
          ;
          if (stickyStart) {
            start -= 1;
          }
          ;
          if (offset > start && offset > start + len) {
            offset += text.length - len;
          }
          ;
        }
        ;
      }
      ;
    } else if (to < from) {
      while (to < from) {
        let edits = this.history[from--];
        for (let φ17 = 0, φ18 = iter$__7(edits), φ19 = φ18.length; φ17 < φ19; φ17++) {
          let [start, len, text] = φ18[φ17];
          if (start > offset) {
            continue;
          }
          ;
          if (offset > start && offset > start + len) {
            offset -= text.length - len;
          }
          ;
        }
        ;
      }
      ;
    }
    ;
    return offset;
  }
  historicalOffset(offset, oldVersion) {
    return this.offsetAtVersion(offset, this.version, oldVersion, true);
  }
  applyEdit(change, version, changes) {
    this.content = this.content.substring(0, change.range.start.offset) + change.text + this.content.substring(change.range.end.offset, this.content.length);
    let line = change.range.start.line;
    this.invalidateFromLine(line);
    return;
  }
  updated(changes, version) {
    this.version = version;
    return this;
  }
  invalidateFromLine(line) {
    this.head = this.seed;
    return this;
  }
  after(token, match) {
    let idx = this.tokens.indexOf(token);
    if (match) {
      while (idx < this.tokens.length) {
        let tok = this.tokens[++idx];
        if (tok && this.matchToken(tok, match)) {
          return tok;
        }
        ;
      }
      ;
      return null;
    }
    ;
    return this.tokens[idx + 1];
  }
  matchToken(token, match) {
    if (match instanceof RegExp) {
      return token.type.match(match);
    } else if (typeof match == "string") {
      return token.type == match;
    }
    ;
    return false;
  }
  before(token, match, offset = 0) {
    let idx = this.tokens.indexOf(token) + offset;
    if (match) {
      while (idx > 0) {
        let tok = this.tokens[--idx];
        if (this.matchToken(tok, match)) {
          return tok;
        }
        ;
      }
      ;
      return null;
    }
    ;
    return this.tokens[idx - 1];
  }
  getTokenRange(token) {
    return {start: this.positionAt(token.offset), end: this.positionAt(token.offset + token.value.length)};
  }
  getTokensInScope(scope) {
    let start = this.tokens.indexOf(scope.start);
    let end = scope.end ? this.tokens.indexOf(scope.end) : this.tokens.length;
    let i = start;
    let parts = [];
    while (i < end) {
      let tok = this.tokens[i++];
      if (tok.scope && tok.scope != scope) {
        parts.push(tok.scope);
        i = tok.scope.endIndex + 1;
      } else {
        parts.push(tok);
      }
      ;
    }
    ;
    return parts;
  }
  getSymbols() {
    var φ21;
    this.astify();
    return (φ21 = this[φ20]).symbols || (φ21.symbols = this.tokens.map(function(_0) {
      return _0.symbol;
    }).filter(function(_0) {
      return _0;
    }).filter(function(sym, i, arr) {
      return arr.indexOf(sym) == i;
    }));
  }
  getImportedSymbols() {
    return this.getSymbols().filter(function(_0) {
      return _0.importedΦ;
    });
  }
  getImportNodes() {
    let tokens = this.tokens.filter(function(_0) {
      return _0.match("push._imports");
    });
    return tokens.map(function(_0) {
      return _0.scope;
    });
  }
  getNodesInScope(scope, includeEnds = false) {
    let tok = scope.start;
    let end = scope.end;
    if (includeEnds) {
      end = end.next;
    } else {
      tok = tok.next;
    }
    ;
    let parts = [];
    while (tok && tok != end) {
      if (tok.scope && tok.scope != scope) {
        parts.push(tok.scope);
        tok = tok.scope.end.next;
        continue;
      } else if (tok.type != "white") {
        parts.push(tok);
      }
      ;
      tok = tok.next;
    }
    ;
    return parts;
  }
  getTokenAtOffset(offset, forwardLooking = false) {
    return this.tokenAtOffset(offset);
    let pos = this.positionAt(offset);
    this.getTokens(pos);
    let line = this.lineTokens[pos.line];
    let idx = line.index;
    let token;
    let prev;
    while (token = this.tokens[idx++]) {
      if (forwardLooking && token.offset == offset) {
        return token;
      }
      ;
      if (token.offset >= offset) {
        break;
      }
      ;
      prev = token;
    }
    ;
    return prev || token;
  }
  getSemanticTokens(filter = SymbolFlags.Scoped) {
    let tokens = this.parse();
    let items = [];
    for (let i = 0, φ222 = iter$__7(tokens), φ23 = φ222.length; i < φ23; i++) {
      let tok = φ222[i];
      let sym = tok.symbol;
      if (!(sym && (!filter || sym.flags & filter))) {
        continue;
      }
      ;
      let typ = SemanticTokenTypes[sym.semanticKind];
      let mods = tok.mods | sym.semanticFlags;
      items.push([tok.offset, tok.value.length, typ, mods]);
    }
    ;
    return items;
  }
  getEncodedSemanticTokens() {
    let tokens = this.getSemanticTokens();
    let out = [];
    let l = 0;
    let c = 0;
    for (let φ24 = 0, φ25 = iter$__7(tokens), φ26 = φ25.length; φ24 < φ26; φ24++) {
      let item = φ25[φ24];
      let pos = this.positionAt(item[0]);
      let dl = pos.line - l;
      let chr = dl ? pos.character : pos.character - c;
      out.push(dl, chr, item[1], item[2], item[3]);
      l = pos.line;
      c = pos.character;
    }
    ;
    return out;
  }
  getDestructuredPath(tok, stack = [], root = null) {
    if (tok.context.type == "array") {
      this.getDestructuredPath(tok.context.start, stack, root);
      stack.push(tok.context.indexOfNode(tok));
      return stack;
    }
    ;
    let key = tok.value;
    if (tok.prev.match("operator.assign.key-value")) {
      key = tok.prev.prev.value;
    }
    ;
    if (tok.context.type == "object") {
      this.getDestructuredPath(tok.context.start, stack, root);
      stack.push(key);
    }
    ;
    return stack;
  }
  tokenAtOffset(offset) {
    let tok = this.tokens[0];
    while (tok) {
      let next = tok.next;
      if (tok.offset >= offset) {
        return tok.prev;
      }
      ;
      if (tok.end && tok.end.offset < offset) {
        tok = tok.end;
      } else if (next) {
        tok = next;
      } else {
        return tok;
      }
      ;
    }
    ;
    return tok;
  }
  patternAtOffset(offset, matcher = /[\w\-\.\%]/) {
    let from = offset;
    let to = offset;
    let str = this.content;
    while (from > 0 && matcher.test(this.content[from - 1])) {
      from--;
    }
    ;
    while (matcher.test(this.content[to + 1] || "")) {
      to++;
    }
    ;
    let value = str.slice(from, to + 1);
    return [value, from, to];
  }
  adjustmentAtOffset(offset, amount = 1) {
    let [word, start, end] = this.patternAtOffset(offset);
    let [pre, post = ""] = word.split(/[\d\.]+/);
    let num = parseFloat(word.slice(pre.length).slice(0, post.length ? -post.length : 1e3));
    if (!Number.isNaN(num)) {
      num += amount;
      return [start + pre.length, word.length - pre.length - post.length, String(num)];
    }
    ;
    return null;
  }
  contextAtOffset(offset) {
    var φ27;
    this.ensureParsed();
    let pos = this.positionAt(offset);
    let tok = this.tokenAtOffset(offset);
    let linePos = this.lineOffsets[pos.line];
    let tokPos = offset - tok.offset;
    let ctx = tok.context;
    const before = {
      character: this.content[offset - 1],
      line: this.content.slice(linePos, offset),
      token: tok.value.slice(0, tokPos)
    };
    const after = {
      character: this.content[offset],
      token: tok.value.slice(tokPos),
      line: this.content.slice(offset, this.lineOffsets[pos.line + 1]).replace(/[\r\n]+/, "")
    };
    if (tok.scope && !after.token) {
      ctx = tok.scope;
    }
    ;
    if (tok.next) {
      if (tok.next.value == null && tok.next.scope && !after.token && tok.match("operator.assign")) {
        ctx = tok.next.scope;
      }
      ;
    }
    ;
    let tabs = prevToken(tok, "white.tabs");
    let indent = tabs ? tabs.value.length : 0;
    let group = ctx;
    let scope = ctx.scope;
    let meta = {};
    let target = tok;
    let mstate = tok.stack.state || "";
    let t = CompletionTypes;
    if (group) {
      if (group.start) {
        before.group = this.content.slice(group.start.offset, offset);
      }
      ;
      if (group.end) {
        after.group = this.content.slice(offset, group.end.offset);
      }
      ;
    }
    ;
    let suggest = {
      keywords: []
    };
    let flags = 0;
    if (tok == tabs) {
      indent = tokPos;
    }
    ;
    if (tok.match("br white.tabs")) {
      while (scope.indent > indent) {
        scope = scope.parent;
      }
      ;
    }
    ;
    if (group.type == "tag") {
      true;
    }
    ;
    if (tok.match("entity string regexp comment style.")) {
      flags = 0;
    }
    ;
    if (tok.match("tag.event.name tag.event-modifier.name")) {
      target = tok.prev;
    }
    ;
    if (tok.type == "path" || tok.type == "path.open") {
      flags |= CompletionTypes.Path;
      suggest.paths = 1;
    }
    ;
    if (tok.match("identifier tag.operator.equals br white delimiter array operator ( self")) {
      flags |= CompletionTypes.Value;
      target = null;
    }
    ;
    if (tok.match("operator.access")) {
      flags |= CompletionTypes.Access;
      target = tok;
    }
    ;
    if (tok.match("accessor")) {
      flags |= CompletionTypes.Access;
      target = tok.prev;
    }
    ;
    if (tok.match("delimiter.type.prefix type")) {
      flags |= CompletionTypes.Type;
    }
    ;
    if (tok.match("tag.name tag.open")) {
      flags |= CompletionTypes.TagName;
    } else if (tok.match("tag.attr tag.white")) {
      flags |= CompletionTypes.TagProp;
    } else if (tok.match("tag.flag")) {
      flags |= CompletionTypes.TagFlag;
    } else if (tok.match("tag.event.modifier")) {
      flags |= CompletionTypes.TagEventModifier;
    } else if (tok.match("tag.event")) {
      flags |= CompletionTypes.TagEvent;
    } else if (tok.match("operator.equals.tagop")) {
      flags |= CompletionTypes.Value;
    }
    ;
    if (tok.match("style.property.operator") || group.closest("stylevalue")) {
      flags |= t.StyleValue;
      try {
        suggest.styleProperty = group.closest("styleprop").propertyName;
      } catch (e) {
      }
      ;
    }
    ;
    if (tok.match("style.open style.property.name")) {
      flags |= t.StyleProp;
    }
    ;
    if (tok.match("style.value.white") || tok.prev && tok.prev.match("style.value.white")) {
      flags |= t.StyleProp;
    }
    ;
    if (tok.match("style.selector.element") && after.line.match(/^\s*$/)) {
      flags |= t.StyleProp;
    }
    ;
    if (scope.closest("rule")) {
      flags |= t.StyleProp;
      flags &= ~t.Value;
    }
    ;
    if (tok.match("style.property.operator")) {
      flags &= ~t.StyleProp;
    }
    ;
    if (group.match("stylevalue") && before.group.indexOf(" ") == -1) {
      flags = t.StyleValue;
    }
    ;
    if (tok.match("style.selector.modifier style.property.modifier")) {
      flags = t.StyleModifier;
    }
    ;
    if (tok.match("style.selector.element")) {
      flags |= t.StyleSelector;
    }
    ;
    if (scope.closest("rule") && before.line.match(/^\s*$/)) {
      flags |= t.StyleSelector;
      flags &= ~t.StyleValue;
    }
    ;
    if (tok.match("operator.access accessor white.classname white.tagname")) {
      flags &= ~t.Value;
    }
    ;
    if (group.closest("imports")) {
      flags &= ~t.Value;
      flags |= t.ImportName;
    }
    ;
    if (mstate.match(/\.decl-(let|var|const|param|for)/) || tok.match(/\.decl-(for|let|var|const|param)/)) {
      flags &= ~t.Value;
      flags |= t.VarName;
    }
    ;
    let kfilter = scope.allowedKeywordTypes;
    φ27 = [];
    for (let φ28 = 0, φ29 = Object.keys(Keywords), φ30 = φ29.length, key, v; φ28 < φ30; φ28++) {
      key = φ29[φ28];
      v = Keywords[key];
      if (!(v & kfilter)) {
        continue;
      }
      ;
      φ27.push(key);
    }
    ;
    suggest.keywords = φ27;
    suggest.flags = flags;
    for (let φ31 = 0, φ32 = Object.keys(t), φ33 = φ32.length, k, v; φ31 < φ33; φ31++) {
      k = φ32[φ31];
      v = t[k];
      if (flags & v) {
        suggest[k] || (suggest[k] = true);
      }
      ;
    }
    ;
    let out = {
      token: tok,
      offset,
      position: pos,
      linePos,
      scope,
      indent,
      group: ctx,
      mode: "",
      target,
      path: scope.path,
      suggest,
      before,
      after
    };
    return out;
  }
  textBefore(offset) {
    let before = this.content.slice(0, offset);
    let ln = before.lastIndexOf("\n");
    return before.slice(ln + 1);
  }
  varsAtOffset(offset, globalsΦ = false) {
    let tok = this.tokenAtOffset(offset);
    let vars = [];
    let scope = tok.context.scope;
    let names = {};
    while (scope) {
      for (let φ34 = 0, φ35 = iter$__7(Object.values(scope.varmap)), φ36 = φ35.length; φ34 < φ36; φ34++) {
        let item = φ35[φ34];
        if (item.globalΦ && !globalsΦ) {
          continue;
        }
        ;
        if (names[item.name]) {
          continue;
        }
        ;
        if (!item.node || item.node.offset < offset) {
          vars.push(item);
          names[item.name] = item;
        }
        ;
      }
      ;
      scope = scope.parent;
    }
    ;
    return vars;
  }
  getOutline(walker = null) {
    var φ39, φ45, φ46, φ47;
    if (this.isLegacy) {
      let symbols2 = fastExtractSymbols(this.content);
      for (let φ37 = 0, φ38 = iter$__7(symbols2.all), φ40 = φ38.length; φ37 < φ40; φ37++) {
        let item = φ38[φ37];
        φ39 = item.parent, delete item.parent, φ39;
        item.path = item.name;
        item.name = item.ownName;
        if (walker) {
          walker(item, symbols2.all);
        }
        ;
      }
      ;
      return symbols2;
    }
    ;
    this.ensureParsed();
    let t = Date.now();
    let all = [];
    let root = {children: []};
    let curr = root;
    let scop = null;
    let last = {};
    let symbols = new Set();
    let awaitScope = null;
    function add(item, tok) {
      if (item instanceof Sym) {
        symbols.add(item);
        item = {
          name: item.name,
          kind: item.kind
        };
      }
      ;
      last = item;
      item.token = tok;
      item.children || (item.children = []);
      item.span || (item.span = tok.span);
      item.name || (item.name = tok.value);
      all.push(item);
      return curr.children.push(item);
    }
    ;
    function push(end) {
      last.children || (last.children = []);
      last.parent || (last.parent = curr);
      curr = last;
      return curr.end = end;
    }
    ;
    function pop(tok) {
      return curr = curr.parent;
    }
    ;
    for (let i = 0, φ41 = iter$__7(this.tokens), φ42 = φ41.length; i < φ42; i++) {
      let token = φ41[i];
      let sym = token.symbol;
      let scope = token.scope;
      if (token.type == "key") {
        add({kind: SymbolKind.Key}, token);
      } else if (sym) {
        if (sym.parameterΦ) {
          continue;
        }
        ;
        if (!symbols.has(sym)) {
          add(sym, token);
        }
        ;
        if (sym.body) {
          awaitScope = sym.body.start;
        }
        ;
      } else if (scope && scope.type == "do") {
        let pre = this.textBefore(token.offset - 3).replace(/^\s*(return\s*)?/, "");
        pre += " callback";
        add({kind: SymbolKind.Function, name: pre}, token.prev);
        awaitScope = token;
      } else if (scope && scope.type == "tag") {
        add({kind: SymbolKind.Field, name: scope.outline}, token);
      }
      ;
      if (token == awaitScope) {
        push(token.end);
      }
      ;
      if (token == curr.end) {
        pop();
      }
      ;
    }
    ;
    for (let φ43 = 0, φ44 = iter$__7(all), φ48 = φ44.length; φ43 < φ48; φ43++) {
      let item = φ44[φ43];
      if (item.span) {
        let len = item.span.length;
        item.span.start = this.positionAt(item.span.offset);
        item.span.end = len ? this.positionAt(item.span.offset + len) : item.span.start;
      }
      ;
      if (walker) {
        walker(item, all);
      }
      ;
      φ45 = item.parent, delete item.parent, φ45;
      φ46 = item.end, delete item.end, φ46;
      φ47 = item.token, delete item.token, φ47;
    }
    ;
    return root;
  }
  getContextAtOffset(offset, forwardLooking = false) {
    return this.contextAtOffset(offset);
  }
  ensureParsed() {
    this.parse();
    return this;
  }
  reparse() {
    this.invalidateFromLine(0);
    return this.parse();
  }
  profileReparse() {
    let t = Date.now();
    let res = this.reparse();
    console.log("took", Date.now() - t);
    return res;
  }
  tokenize(force = false) {
    var φ51;
    let from = this[φ20] || {lines: [], version: -1};
    if (from.version == this.version && !force) {
      return from;
    }
    ;
    let raw = this.content;
    if (this.isLegacy) {
      raw = raw.replace(/\@\w/g, function(m) {
        return "¶" + m.slice(1);
      });
      raw = raw.replace(/\w\:(?=\w)/g, function(m) {
        return m[0] + ".";
      });
      raw = raw.replace(/(do)(\s?)\|([^\|]*)\|/g, function(m, a, space, b) {
        return a + "(" + (space || "") + b + ")";
      });
    }
    ;
    let lineOffsets = this.lineOffsets;
    let tokens = [];
    let head = this.seed;
    let prev = head;
    let t0 = Date.now();
    let nextState = this.initialState;
    this[φ20] = {
      version: this.version,
      lines: [],
      tokens
    };
    let lineCache = {};
    this[φ20].cache = lineCache;
    for (let φ49 = 0, φ50 = iter$__7(from.lines), φ52 = φ50.length; φ49 < φ52; φ49++) {
      let line = φ50[φ49];
      let item = lineCache[φ51 = line.text] || (lineCache[φ51] = []);
      item.push(line);
      line;
    }
    ;
    for (let i = 0, φ53 = iter$__7(lineOffsets), φ56 = φ53.length; i < φ56; i++) {
      let lineOffset = φ53[i];
      let next = lineOffsets[i + 1];
      let toOffset = next || raw.length;
      let str = raw.slice(lineOffset, toOffset);
      let startState = nextState;
      let cached = lineCache[str];
      let matches = cached && cached.filter(function(item) {
        return item.startState == startState;
      });
      let match = matches && (matches.find(function(_0) {
        return _0.offset == lineOffset;
      }) || matches[0]);
      let lexed = null;
      if (match) {
        if (match.offset == lineOffset) {
          lexed = match.clone(lineOffset);
        } else {
          lexed = match.clone(lineOffset);
        }
        ;
      }
      ;
      if (!lexed) {
        let run = lexer.tokenize(str, startState, lineOffset);
        lexed = new LexedLine({
          offset: lineOffset,
          text: str,
          startState,
          endState: run.endState,
          tokens: run.tokens
        });
      }
      ;
      for (let ti = 0, φ54 = iter$__7(lexed.tokens), φ55 = φ54.length; ti < φ55; ti++) {
        let tok = φ54[ti];
        tokens.push(tok);
      }
      ;
      this[φ20].lines.push(lexed);
      nextState = lexed.endState;
    }
    ;
    return this[φ20];
  }
  get tokens() {
    this.astify();
    return this[φ20].tokens;
  }
  getTokens(range = null) {
    return this.tokens;
  }
  astify() {
    let lexed = this.tokenize();
    if (lexed.root) {
      return this;
    }
    ;
    const pairings = {"]": "[", ")": "(", "}": "{", ">": "<"};
    const openers = {"[": "]", "(": ")", "{": "}", "<": ">"};
    const callAfter = /[\w\$\)\]\?]/;
    let t0 = Date.now();
    let entity = null;
    let scope = lexed.root = new Root(this, this.seed, null, "root");
    let raw = this.content;
    let log2 = console.log.bind(console);
    let lastDecl = null;
    let lastVarKeyword = null;
    let lastVarAssign = null;
    let prev = null;
    let entityFlags = 0;
    for (let ti = 0, φ57 = iter$__7(lexed.tokens), φ58 = φ57.length; ti < φ58; ti++) {
      let tok = φ57[ti];
      let types = tok.type.split(".");
      let value = tok.value;
      let nextToken = lexed.tokens[ti + 1];
      let [typ, subtyp, sub2] = types;
      let ltyp = types[types.length - 1];
      let styp = types[types.length - 2];
      let scopeType = null;
      let decl = 0;
      if (typ == "ivar") {
        value = tok.value = "@" + value.slice(1);
      }
      ;
      if (prev) {
        prev.next = tok;
      }
      ;
      tok.prev = prev;
      tok.context = scope;
      if (typ == "(" && prev) {
        let prevchr = raw[tok.offset - 1] || "";
        if (callAfter.test(prevchr)) {
          scope = tok.scope = ScopeTypeMap.args.build(this, tok, scope, "args", types);
        }
        ;
      }
      ;
      if (typ == "operator") {
        tok.op = tok.value.trim();
      }
      ;
      if (typ == "keyword") {
        if (M[subtyp]) {
          entityFlags |= M[subtyp];
        }
        ;
        if (value == "let" || value == "const") {
          lastVarKeyword = tok;
          lastVarAssign = null;
        }
        ;
      }
      ;
      if (typ == "entity") {
        tok.mods |= entityFlags;
        entityFlags = 0;
      }
      ;
      if (typ == "push") {
        let scopetype = subtyp;
        let idx = subtyp.lastIndexOf("_");
        let ctor = idx >= 0 ? Group : Scope;
        if (idx >= 0) {
          scopetype = scopetype.slice(idx + 1);
          ctor = ScopeTypeMap[scopetype] || Group;
        } else if (ScopeTypeMap[scopetype]) {
          ctor = ScopeTypeMap[scopetype];
        }
        ;
        scope = tok.scope = new ctor(this, tok, scope, scopetype, types);
        if (lastDecl) {
          lastDecl.body = scope;
          scope.symbol = lastDecl;
          lastDecl = null;
        }
        ;
        if (scope == scope.scope) {
          lastVarKeyword = null;
          lastVarAssign = null;
        }
        ;
      } else if (typ == "pop") {
        if (subtyp == "value") {
          lastVarAssign = null;
        }
        ;
        scope = scope.pop(tok);
      } else if ((subtyp == "open" || openers[subtyp]) && ScopeTypeMap[typ]) {
        scope = tok.scope = ScopeTypeMap[typ].build(this, tok, scope, typ, types);
      } else if (ltyp == "open" && (scopeType = ScopeTypeMap[styp])) {
        scope = tok.scope = scopeType.build(this, tok, scope, styp, types);
      } else if (ltyp == "close" && scope.type == styp) {
        scope = scope.pop(tok);
      } else if (subtyp == "close" && ScopeTypeMap[typ]) {
        scope = scope.pop(tok);
      } else if (pairings[typ] && scope && scope.start.value == pairings[typ]) {
        scope = scope.pop(tok);
      }
      ;
      if (tok.match(/entity\.name|decl-/)) {
        let tokenSymbol = Sym.forToken(tok, tok.type, tok.mods);
        if (tokenSymbol) {
          lastDecl = tok.symbol = tokenSymbol;
          tok.symbol.keyword = lastVarKeyword;
          scope.register(tok.symbol);
        }
        ;
        tok.mods |= M.Declaration;
      }
      ;
      if (subtyp == "declval") {
        lastVarAssign = tok;
      }
      ;
      if (tok.match("identifier") && !tok.symbol) {
        let sym = scope.lookup(tok, lastVarKeyword);
        if (sym && sym.scopedΦ) {
          if (lastVarAssign && sym.keyword == lastVarKeyword) {
            true;
          } else {
            sym.addReference(tok);
          }
          ;
        }
        ;
        if (prev && prev.op == "=" && sym) {
          let lft = prev.prev;
          if (lft && lft.symbol == sym) {
            if (lft.mods & M.Declaration) {
              sym.dereference(tok);
            } else if (!nextToken || nextToken.match("br")) {
              sym.dereference(lft);
            }
            ;
          }
          ;
        }
        ;
      }
      ;
      prev = tok;
    }
    ;
    return this;
  }
  parse() {
    return this.tokens;
  }
  getMatchingTokens(filter) {
    let tokens = this.getTokens();
    tokens = tokens.slice(0).filter(function(_0) {
      return _0.match(filter);
    });
    return tokens;
  }
  migrateToImba2() {
    let source = this.content;
    source = source.replace(/\bdef self\./g, "static def ");
    source = source.replace(/\b(var|let|const) def /g, "def ");
    source = source.replace(/\?\./g, "..");
    source = source.replace(/def ([\w\-]+)\=/g, "set $1");
    source = source.replace(/do\s?\|([^\|]+)\|/g, "do($1)");
    source = source.replace(/(prop) ([\w\-]+) (.+)$/gm, function(m, typ, name, rest) {
      var φ59, φ60;
      let opts = {};
      rest.split(/,\s*/).map(function(_0) {
        return _0.split(/\:\s*/);
      }).map(function(_0) {
        return opts[_0[0]] = _0[1];
      });
      let out = "" + typ + " " + name;
      if (opts.watch && opts.watch[0].match(/[\'\"\:]/)) {
        out = "@watch(" + opts.watch + ") " + out;
      } else if (opts.watch) {
        out = "@watch " + out;
      }
      ;
      φ59 = opts.watch, delete opts.watch, φ59;
      if (opts.default) {
        out = "" + out + " = " + opts.default;
        φ60 = opts.default, delete opts.default, φ60;
      }
      ;
      if (Object.keys(opts).length) {
        console.log("more prop values", m, opts);
      }
      ;
      return out;
    });
    let doc = ImbaDocument.tmp(source);
    let tokens = doc.getTokens();
    let ivarPrefix = "";
    for (let i = 0, φ61 = iter$__7(tokens), φ62 = φ61.length; i < φ62; i++) {
      let token = φ61[i];
      let next = tokens[i + 1];
      let {value, type, offset} = token;
      let end = offset + value.length;
      if (type == "operator.dot.legacy") {
        value = ".";
        if (next) {
          next.access = true;
        }
        ;
      }
      ;
      if (type == "operator.spread.legacy") {
        value = "...";
      }
      ;
      if (type == "identifier.tagname") {
        if (value.indexOf(":") >= 0) {
          value = value.replace(":", "-");
        }
        ;
      }
      ;
      if (type == "identifier.def.propname" && value == "initialize") {
        value = "constructor";
      }
      ;
      if (type == "decorator" && !source.slice(end).match(/^\s(prop|def|get|set)/)) {
        value = ivarPrefix + value.slice(1);
      }
      ;
      if (type == "property") {
        if (value[0] == "@") {
          value = value.replace(/^\@/, ivarPrefix);
          token.access = true;
        } else if (value == "len") {
          value = "length";
        } else if (/^(\n|\s\:|\)|\,|\.)/.test(source.slice(end)) && !token.access) {
          if (value[0] == value[0].toLowerCase()) {
            value = value + "!";
          }
          ;
        }
        ;
      }
      ;
      if (type == "identifier" && !token.access && value[0] == value[0].toLowerCase() && value[0] != "_") {
        if (!token.variable && /^(\n|\s\:|\)|\,|\.)/.test(source.slice(end)) && value != "new") {
          value = value + "!";
        }
        ;
      }
      ;
      token.value = value;
    }
    ;
    return tokens.map(function(_0) {
      return _0.value;
    }).join("");
  }
  createImportEdit(path, name, alias = name) {
    path = path.replace(/\.imba$/, "");
    let nodes = this.getImportNodes().filter(function(_0) {
      return _0.sourcePath == path;
    });
    let out = "";
    let offset = 0;
    let changes = [];
    let result = {
      changes
    };
    if (true) {
      let symbols = this.getImportedSymbols().map(function(_0) {
        return _0.importInfo;
      });
      let match = symbols.find(function(_0) {
        return _0.path == path && _0.name == alias && _0.exportName == name;
      });
      if (match) {
        return result;
      }
      ;
    }
    ;
    if (name != "default" && name != "*") {
      nodes = nodes.filter(function(_0) {
        return _0.specifiers || !_0.ns;
      });
    }
    ;
    for (let φ63 = 0, φ64 = iter$__7(nodes), φ65 = φ64.length; φ63 < φ65; φ63++) {
      let node = φ64[φ63];
      let defaults = node.default;
      let members = node.specifiers;
      let ns = node.namespace;
      if (name == "default") {
        offset = node.start.offset + 1;
        if (defaults) {
          if (defaults.value == alias) {
            return result;
          } else {
            result.alias = defaults.value;
            offset = 0;
            continue;
          }
          ;
        } else {
          out = alias;
          if (ns || members) {
            out += ", ";
          }
          ;
        }
        ;
      } else if (name == "*") {
        if (members) {
          continue;
        }
        ;
        if (defaults) {
          offset = defaults.endOffset;
          out = ", * as " + alias;
        } else {
          offset = node.start.offset + 1;
          out = "* as " + alias + " ";
        }
        ;
      } else if (ns) {
        result.alias = "" + ns.value + "." + name;
        continue;
      } else {
        let key = name;
        if (alias != name) {
          key += " as " + alias;
        }
        ;
        if (members) {
          offset = members.start.offset + 1;
          out = " " + key + ",";
        } else if (defaults) {
          offset = defaults.endOffset;
          out = ", { " + key + " }";
        } else {
          out = "{ " + key + " }";
          offset = node.start.offset + 1;
        }
        ;
      }
      ;
      if (out) {
        break;
      }
      ;
    }
    ;
    if (!out) {
      if (name == "default") {
        out = "import " + alias + " from '" + path + "'";
      } else if (name == "*") {
        out = "import * as " + alias + " from '" + path + "'";
      } else if (alias != name) {
        out = "import { " + name + " as " + alias + " } from '" + path + "'";
      } else {
        out = "import { " + name + " } from '" + path + "'";
      }
      ;
      out += "\n";
    }
    ;
    changes.push({newText: out, range: this.rangeAt(offset, offset)});
    return result;
  }
};

// src/program/grammars/xml.imba
var grammar2 = {
  defaultToken: "",
  tokenPostfix: ".xml",
  ignoreCase: true,
  qualifiedName: /(?:[\w\.\-]+:)?[\w\.\-]+/,
  tokenizer: {
    root: [
      [/[^<&]+/, ""],
      {include: "@whitespace"},
      [/(<)(@qualifiedName)/, [
        {token: "delimiter"},
        {token: "tag", next: "@tag"}
      ]],
      [/(<\/)(@qualifiedName)(\s*)(>)/, [
        {token: "delimiter"},
        {token: "tag"},
        "",
        {token: "delimiter"}
      ]],
      [/(<\?)(@qualifiedName)/, [
        {token: "delimiter"},
        {token: "metatag", next: "@tag"}
      ]],
      [/(<\!)(@qualifiedName)/, [
        {token: "delimiter"},
        {token: "metatag", next: "@tag"}
      ]],
      [/<\!\[CDATA\[/, {token: "delimiter.cdata", next: "@cdata"}],
      [/&\w+;/, "string.escape"]
    ],
    cdata: [
      [/[^\]]+/, ""],
      [/\]\]>/, {token: "delimiter.cdata", next: "@pop"}],
      [/\]/, ""]
    ],
    tag: [
      [/[ \t\r\n]+/, ""],
      [/(@qualifiedName)(\s*=\s*)("[^"]*"|'[^']*')/, ["attribute.name", "", "attribute.value"]],
      [/(@qualifiedName)(\s*=\s*)("[^">?\/]*|'[^'>?\/]*)(?=[\?\/]\>)/, ["attribute.name", "", "attribute.value"]],
      [/(@qualifiedName)(\s*=\s*)("[^">]*|'[^'>]*)/, ["attribute.name", "", "attribute.value"]],
      [/@qualifiedName/, "attribute.name"],
      [/\?>/, {token: "delimiter", next: "@pop"}],
      [/(\/)(>)/, [
        {token: "tag"},
        {token: "delimiter", next: "@pop"}
      ]],
      [/>/, {token: "delimiter", next: "@pop"}]
    ],
    whitespace: [
      [/[ \t\r\n]+/, ""],
      [/<!--/, {token: "comment", next: "@comment"}]
    ],
    comment: [
      [/[^<\-]+/, "comment.content"],
      [/-->/, {token: "comment", next: "@pop"}],
      [/<!--/, "comment.content.invalid"],
      [/[<\-]/, "comment.content"]
    ]
  }
};

// src/program/monarch.imba
var tokenizers = {};
var Monarch = class {
  static getTokenizer(langId) {
    if (langId == "xml" && !tokenizers[langId]) {
      return this.createTokenizer("xml", grammar2);
    }
    ;
    return tokenizers[langId];
  }
  static createTokenizer(langId, grammar3) {
    let compiled2 = compile(langId, grammar3);
    return tokenizers[langId] = new MonarchTokenizer(langId, compiled2);
  }
};

// src/program/highlighter.imba
function iter$__8(a) {
  let v;
  return a ? (v = a.toIterable) ? v.call(a) : a : [];
}
var replacements = {
  "&": "&amp;",
  "<": "&lt;",
  ">": "&gt;",
  '"': "&quot;",
  "'": "&#39;"
};
var typenames = {
  "[": "square open",
  "]": "square close",
  "{": "curly open",
  "}": "curly close",
  "(": "paren open",
  ")": "paren close"
};
function escape(str) {
  return str.replace(/[\&\<\>]/g, function(m) {
    return replacements[m];
  });
}
function classify(types) {
  return types.join(" ").replace(/[\[\]\{\}\(\)]/g, function(m) {
    return typenames[m];
  });
}
function highlight(tokens, {verbose = false}) {
  let parts = [];
  let depth = 0;
  let counter = 0;
  let ids = [];
  for (let φ16 = 0, φ23 = iter$__8(tokens), φ32 = φ23.length; φ16 < φ32; φ16++) {
    let token = φ23[φ16];
    let value = token.value;
    let types = token.type.split(".");
    let [typ, subtyp] = types;
    let mods = token.mods;
    if (token.var) {
      let id = ids.indexOf(token.var);
      if (id == -1) {
        id = ids.push(token.var) - 1;
      }
      ;
      types.push("vref");
      types.push("var" + id);
      types.push(token.var.type + "-ref");
      mods |= token.var.mods;
    }
    ;
    if (mods & M.Declaration) {
      types.push("decl");
    }
    ;
    if (mods & M.Root) {
      types.push("root");
    }
    ;
    if (mods & M.Local) {
      types.push("local");
    }
    ;
    if (token.end && token.end.start == token) {
      parts.push("<b>");
    }
    ;
    if (subtyp == "start" || subtyp == "open") {
      parts.push("<b class='" + typ + "'>");
      if (!value) {
        continue;
      }
      ;
    }
    ;
    if ((subtyp == "end" || subtyp == "close") && !value) {
      parts.push("</b>");
      continue;
    }
    ;
    if (typ == "push") {
      value = String(++depth);
      let kind = subtyp.indexOf("_") >= 0 ? "group" : "scope";
      let end = token.scope && token.scope.end;
      parts.push("<div class='" + kind + "-" + subtyp.split("_").pop() + " _" + subtyp + " l" + depth + " o" + token.offset + " e" + (end && end.offset) + "'>");
      continue;
    } else if (typ == "pop") {
      value = String(--depth);
      parts.push("</div>");
      continue;
    }
    ;
    if (typ != "white" && typ != "line") {
      value = "<i class='" + classify(types) + " o" + token.offset + "'>" + escape(value || "") + "</i>";
    } else if (typ == "white" && value != "\n") {
      value = "<i raw='" + JSON.stringify(value) + "'>" + escape(value || "") + "</i>";
    }
    ;
    parts.push(value);
    if (subtyp == "end" || subtyp == "close" || token.start) {
      parts.push("</b>");
    }
    ;
  }
  ;
  return parts.join("");
}
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  CompletionTypes,
  Group,
  ImbaDocument,
  KeywordTypes,
  Keywords,
  M,
  Monarch,
  Node,
  Position,
  Range,
  Scope,
  SemanticTokenModifiers,
  SemanticTokenTypes,
  Sym,
  SymbolFlags,
  SymbolKind,
  Token,
  grammar,
  highlight,
  lexer
});
